#!/usr/bin/env python
# coding: utf-8


# nazvy proměnných jsou v angličtině
# Ale komenty teda ne)


import numpy as np
from . import mart
from . import maxes
from . import maxes3d
from matplotlib import ticker


__all__ = [
            'convergence_diagram', 'convergence_legend', 'double_proxy_diagram', 
            'four_branch_convergence', 'beta_diagram', 'double_diagram', 'suspension_3d', 
            'double_tri_R_plot', 'double_tri_R_twins_plot', 'double_100_plot', 'G_U_distortion',
            'double_pprod_R_plot', 'double_mprod_R_plot',
            'triple_vector_plot',
            'triple_plot', 'triple_wide', 'triple_wide_50_100', 'triple_plot_50_100', 
            'triple_nodes_plot', 'triple_wide_25_50',
            'triple_plot_35_50', 'convex_plot_7', 'quadruple_plot',
            'qhull_under_density', 'plane_under_density', 'dhull_vs_complete'
            ]

space_labels = {
    'R': r'$\mathcal{R}$',
    'Rn': r'$\mathcal{R}^{\mathrm{n}}$',
    'P': r'$\mathcal{P}$',
    'GK': r'$\mathcal{G}^{\mathrm{c}}$',
    'G': r'$\mathcal{G}$',
    'U': r'$\mathcal{U}$'
    }

tri_kwargs = {
    1: dict(ms=3, mew=0.6, lw=0.7, linewidths=[0.7, 0.5, 0.4, 0.3, 0.2, 0.1]),
    32: dict(ms=2.5, mew=0.5, lw=0.7, linewidths=[0.7, 0.5, 0.4, 0.3, 0.2, 0.1]),
    2: dict(ms=2.5, mew=0.4, lw=0.7, linewidths=[0.7, 0.5, 0.4, 0.3, 0.2, 0.1]),
    3: dict(ms=2.2, mew=0.1, lw=0.5, linewidths=np.array([0.7, 0.5, 0.4, 0.3, 0.2, 0.1]) / 1.5)
    }


def suspension_3d(fig, sample_box, space):
    ax3d = fig.add_subplot(121,  projection='3d', azim=-90.0001, elev=89)
    ax3d.space = space
    ax3d.sample_box = sample_box[0:300]
    maxes3d.tri_plot(ax3d)
    ax3d.set_zticks([])
    
    ax = fig.add_subplot(122)
    ax.sample_box = sample_box
    maxes.convergence_square(ax)
    fig.set_figwidth(19/2.54)


def beta_diagram(fig, sample_box, space, lim=100):
    fig.set_figheight(2)
    ax = fig.add_subplot(111)
    ax.sample_box = sample_box
    maxes.beta_diagram(ax)
    #ax.set_ylim(1e-5, 1)
    #ax.set_xlim(0, lim)
    ticks = [0, 80, 160, 240, 320, 400, 480]
    ax.xaxis.set_major_locator(ticker.FixedLocator(locs=ticks))


def convergence_diagram(fig, sample_box, space, lim=1000):
    fig.set_figheight(1.8)
    ax = fig.add_subplot(111)
    ax.sample_box = sample_box
    maxes.convergence_diagram(ax)
    ax.set_xlim(0, lim)

def convergence_legend(fig, sample_box, space, lim=1000):
    fig.set_figheight(2)
    ax = fig.add_subplot(111)
    ax.sample_box = sample_box
    maxes.convergence_legend(ax)
    #ax.legend(bbox_to_anchor=(0.5, -0.25), ncol=2, loc='upper center')
    ax.set_xlim(0, lim)

def four_branch_convergence(fig, sample_box, space, lim=100):
    fig.set_figheight(2)
    ax = fig.add_subplot(111)
    ax.sample_box = sample_box
    maxes.convergence_legend(ax)
    #ax.legend(bbox_to_anchor=(0.5, -0.25), ncol=2, loc='upper center')
    ax.set_ylim(1e-5, 1)
    ax.set_xlim(0, lim)
    seq = ["0", r"$\bm{4}$", r"$\bm{26}$", "50", "75", "100"]
    ticks = [0, 4, 26, 50, 75, 100]
    ax.xaxis.set_major_formatter(ticker.FixedFormatter(seq))
    ax.xaxis.set_major_locator(ticker.FixedLocator(locs=ticks))

def double_diagram(fig, sample_box, space, lim=1000):
    ax1 = ax = fig.add_subplot(211)
    ax.sample_box = sample_box
    maxes.convergence_diagram(ax)
    ax.set_xlim(0, lim)
    ax.set_xlabel('')
    
    ax = fig.add_subplot(212)
    ax.sample_box = sample_box
    maxes.convergence_diagram(ax)
    #ax.sharex(ax1)
    
def double_proxy_diagram(fig, sample_box, space, lim=1000):
    ax1 = ax = fig.add_subplot(211)
    ax.sample_box = sample_box
    maxes.convergence_diagram(ax)
    ax.set_xlim(0, lim)
    ax.set_xlabel('')
    
    ax = fig.add_subplot(212)
    ax.sample_box = sample_box
    maxes.convergence_diagram(ax, sources=['box'], apply_proxy=True)
    ax.sharex(ax1)

def convex_plot_7(fig, sample_box, space):
    ax1 = ax = fig.add_subplot(121)
    ax.sample_box = sample_box[0:7]
    ax.space = space
    
    maxes.convex_hull_plot(ax, tri_space=None, linewidths=[0.7, 0.5, 0.4, 0.3])
    #ax.set_title(space_labels[space], pad=10)
    ax = fig.add_subplot(122)
    ax.sample_box = sample_box
    ax.space = space
    maxes.convex_hull_plot(ax, tri_space=None, linewidths=[0.7, 0.5, 0.4, 0.3])
    #ax.set_title(space_labels[space], pad=10)
    ax.sharey(ax1)


def double_tri_R_twins_plot(fig, sample_box, space):
    ax1 = ax = fig.add_subplot(121)
    ax.sample_box = sample_box
    ax.space = 'R'
    try:
        tn_scheme = sample_box.Tri.tn_scheme
    except:
        import quadpy
        tn_scheme = quadpy.tn.grundmann_moeller(sample_box.nvar, 5)
        
    maxes.tri_nodes_plot(ax, tri_space='R', tn_scheme=tn_scheme)
    ax.set_title(r'$\mathcal{R}$', pad=10)
    ax = fig.add_subplot(122)
    ax.sample_box = sample_box
    ax.space = 'G'
    maxes.tri_nodes_plot(ax, tri_space='R', tn_scheme=tn_scheme)
    ax.set_title(r'$\mathcal{G}$', pad=10)
    y_bound = ax.get_ybound()
    ax1.sharey(ax)
    ax.set_ybound(y_bound)

def double_pprod_R_plot(fig, sample_box, space):
    ax1 = ax = fig.add_subplot(121)
    ax.sample_box = sample_box
    ax.space = 'R'
    try:
        tn_scheme = sample_box.Tri.tn_scheme
    except:
        import quadpy
        tn_scheme = quadpy.tn.grundmann_moeller(sample_box.nvar, 5)
    #linewidths = np.array([0.7, 0.5, 0.4, 0.3, 0.2, 0.1]) / 2
    maxes.tri_nodes_plot(ax, tri_space='R', tn_scheme=tn_scheme, nrid=2000,
                data_offset=0.95, **tri_kwargs[2])
    ax.set_title(space_labels[space], pad=10)
    ax.set_xlim(0, 34)
    ax.set_ylim(0, 34)
    ax = fig.add_subplot(122)
    ax.sample_box = sample_box
    ax.space = 'G'
    maxes.tri_nodes_plot(ax, tri_space='R', tn_scheme=tn_scheme, nrid=2000,
                data_offset=0.7, **tri_kwargs[3])
                #ms=1.5, mew=0.4, lw=0.35, linewidths=linewidths)
    from scipy import stats
    beta = -stats.norm.ppf(ax.sample_box.pf_exact) * np.sqrt(2) #č beta-nebeta
    ax.plot([-beta*5/4, beta/4], [beta/4, -beta*5/4], '-b', lw=0.5, zorder=1150)
    ax.xaxis.set_major_locator(ticker.FixedLocator(locs=[-6]))
    ax.yaxis.set_major_locator(ticker.FixedLocator(locs=[-6]))
    ax.set_title(space_labels['G'], pad=10)


def double_mprod_R_plot(fig, sample_box, space):
    ax1 = ax = fig.add_subplot(121)
    ax.sample_box = sample_box
    ax.space = 'R'
    try:
        tn_scheme = sample_box.Tri.tn_scheme
    except:
        import quadpy
        tn_scheme = quadpy.tn.grundmann_moeller(sample_box.nvar, 5)
    #linewidths = np.array([0.7, 0.5, 0.4, 0.3, 0.2, 0.1]) / 2
    maxes.tri_nodes_plot(ax, tri_space='R', tn_scheme=tn_scheme, nrid=2000,
                data_offset=100, **tri_kwargs[2])
    
    ax.set_title(space_labels[space], pad=10)
    ax.set_xlim(0, 3800)
    ax.set_ylim(0, 3800)
    ax = fig.add_subplot(122)
    ax.sample_box = sample_box
    ax.space = 'G'
    maxes.tri_nodes_plot(ax, tri_space='R', tn_scheme=tn_scheme, nrid=2000,
                data_offset=0.7, **tri_kwargs[3])
                #ms=1.5, mew=0.4, lw=0.35, linewidths=linewidths)
    from scipy import stats
    beta = -stats.norm.ppf(ax.sample_box.pf_exact) * np.sqrt(2) #č beta-nebeta
    ax.plot([-beta/4, beta*5/4], [beta*5/4, -beta/4], '-b', lw=0.5, zorder=1150)
    ax.set_title(space_labels['G'], pad=10)


def double_tri_R_plot(fig, sample_box, space):
    ax1 = ax = fig.add_subplot(121)
    ax.sample_box = sample_box
    ax.space = 'R'
    try:
        tn_scheme = sample_box.Tri.tn_scheme
    except:
        import quadpy
        tn_scheme = quadpy.tn.grundmann_moeller(sample_box.nvar, 5)
        
    linewidths = np.array([0.7, 0.5, 0.4, 0.3, 0.2, 0.1]) / 2
    maxes.tri_nodes_plot(ax, tri_space='R', tn_scheme=tn_scheme, nrid=2000,
                ms=1.5, mew=0.4, lw=0.35, linewidths=linewidths)
    ax.set_title(space_labels[space], pad=10)
    ax = fig.add_subplot(122)
    ax.sample_box = sample_box
    ax.space = 'G'
    maxes.tri_nodes_plot(ax, tri_space='R', tn_scheme=tn_scheme, nrid=2000,
                ms=1.5, mew=0.4, lw=0.35, linewidths=linewidths)
    ax.set_title(space_labels['G'], pad=10)
    
    
    
    
def G_U_distortion(fig, sample_box, space):
    try:
        tn_scheme = sample_box.Tri.tn_scheme
    except:
        import quadpy
        tn_scheme = quadpy.tn.grundmann_moeller(sample_box.nvar, 5)
    ms=3
    mew=0.6
    lw=0.7
    linewidths=[0.7, 0.5, 0.4, 0.3, 0.2, 0.1]
    nrid=100
    
    from .. import simplex as six
    
    
    def _draw_nodes(*args, **kwargs):
        # callback's signature: sx, indices=, simplex=, nodes=, cell_stats=
        # positional "sx" is Tri object itself
        # "indices" are numbers of simplex vertices
        # "simplex" are vertices itself
        # "nodes" is what we really want to draw
        
        event = kwargs['cell_stats']['event']
        
        if event == 'mix':
            color = (149/255, 173/255, 1, 0.7)
            mart.plot_sample(ax1, kwargs['nodes'], ls='', marker='.',\
                         mew=0, mfc=color, ms=2.8, alpha=0.5, rasterized=True)
            mart.plot_sample(ax, kwargs['nodes'], ls='', marker='.',\
                         mew=0, mfc=color, ms=2.8, alpha=0.5, rasterized=True)
        
    
    Tri = six.JustCubatureTriangulation(sample_box, tn_scheme=tn_scheme, \
                                    tri_space='U', issi=None,\
                                    weighting_space=None, \
                                    incremental=False,\
                                    on_add_simplex=_draw_nodes,\
                                    on_delete_simplex=None)
    
    
    ax1 = ax = fig.add_subplot(121)
    ax.sample_box = sample_box
    ax.space = 'G'
    ax.set_aspect(1)
    ax.set_box_aspect(1)
    ax.margins(0)
    ax.set_title(space_labels[ax.space])#, pad=10)
    x_label="$x_1$"; y_label="$x_2$"
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    mart.tri_plot(ax, Tri=Tri, color="#B2B2B2", lw=lw/1.4, zorder=100, rasterized=True)
    mart.plot_points(ax, ms=ms, mew=mew, zorder=100500)
    mart.number_points(ax, ha="left", va="bottom", zorder=1005000)
    try:
        mart.plot_boundaries(ax, lw=lw, zorder=1050, nrod=nrid)
    except:
        pass
    
    ax = fig.add_subplot(122)
    ax.sample_box = sample_box
    ax.space = 'U'
    ax.set_aspect(1)
    ax.set_box_aspect(1)
    ax.margins(0)
    ax.set_title(space_labels[ax.space])#, pad=10)
    ax.set_xlabel(x_label)
    #ax.set_ylabel(y_label)
    mart.triplot(ax, color="#B2B2B2", lw=lw/1.4, zorder=100)
    mart.plot_points(ax, ms=ms, mew=mew, zorder=100500)
    mart.number_points(ax, ha="left", va="bottom", zorder=1005000)
    try:
        mart.plot_boundaries(ax, lw=lw, zorder=1050, nrod=nrid)
    except:
        pass
        
    Tri.integrate()
    
    
    

def double_100_plot(fig, sample_box, space, lim=100):
    #fig.set_figwidth(19/2.54)
    ax = fig.add_subplot(122)
    ax.sample_box = sample_box
    ax.space = space
    try:
        tn_scheme = sample_box.Tri.tn_scheme
    except:
        import quadpy
        tn_scheme = quadpy.tn.grundmann_moeller(sample_box.nvar, 5)
    maxes.tri_nodes_plot(ax, tri_space=None, tn_scheme=tn_scheme, **tri_kwargs[2])
    ax.xaxis.set_major_locator(ticker.FixedLocator(locs=[-6]))
    ax.yaxis.set_major_locator(ticker.FixedLocator(locs=[-6]))
    #ax.set_title(space_labels[space], pad=10)
    ax = fig.add_subplot(121, sharex=ax, sharey=ax)
    ax.sample_box = sample_box[0:lim]
    ax.space = space
    maxes.tri_nodes_plot(ax, tri_space=None, tn_scheme=tn_scheme, **tri_kwargs[2])
    #ax.set_title(space_labels[space], pad=10)
    ax.xaxis.set_major_locator(ticker.FixedLocator(locs=[-6]))
    ax.yaxis.set_major_locator(ticker.FixedLocator(locs=[-6]))


def quadruple_plot(fig, sample_box, space, npoints=[50, 100, 200]):
    #fig.set_figwidth(19/2.54)
    ax = fig.add_subplot(224)
    ax.sample_box = sample_box
    ax.space = space
    try:
        tn_scheme = sample_box.Tri.tn_scheme
    except:
        import quadpy
        tn_scheme = quadpy.tn.grundmann_moeller(sample_box.nvar, 5)
    maxes.tri_nodes_plot(ax, tri_space=None, tn_scheme=tn_scheme, **tri_kwargs[2])
    ax.xaxis.set_major_locator(ticker.FixedLocator(locs=[-6]))
    ax.yaxis.set_major_locator(ticker.FixedLocator(locs=[-6]))
    #ax.set_title(space_labels[space], pad=10)
    ax = fig.add_subplot(223, sharex=ax, sharey=ax)
    ax.sample_box = sample_box[0:npoints[2]]
    ax.space = space
    maxes.tri_nodes_plot(ax, tri_space=None, tn_scheme=tn_scheme, **tri_kwargs[2])
    #ax.set_title(space_labels[space], pad=10)
    ax.xaxis.set_major_locator(ticker.FixedLocator(locs=[-6]))
    ax.yaxis.set_major_locator(ticker.FixedLocator(locs=[-6]))
    ax = fig.add_subplot(222, sharex=ax, sharey=ax)
    ax.sample_box = sample_box[0:npoints[1]]
    ax.space = space
    maxes.tri_nodes_plot(ax, tri_space=None, tn_scheme=tn_scheme, **tri_kwargs[2])
    #ax.set_title(space_labels[space], pad=10)
    ax.xaxis.set_major_locator(ticker.FixedLocator(locs=[-6]))
    ax.yaxis.set_major_locator(ticker.FixedLocator(locs=[-6]))
    ax = fig.add_subplot(221, sharex=ax, sharey=ax)
    ax.sample_box = sample_box[0:npoints[0]]
    ax.space = space
    maxes.tri_nodes_plot(ax, tri_space=None, tn_scheme=tn_scheme, **tri_kwargs[2])
    #ax.set_title(space_labels[space], pad=10)
    ax.xaxis.set_major_locator(ticker.FixedLocator(locs=[-6]))
    ax.yaxis.set_major_locator(ticker.FixedLocator(locs=[-6]))
    

def triple_wide(fig, sample_box, space):
    triple_plot(fig, sample_box, space)
    fig.set_figwidth(19/2.54)

def triple_wide_50_100(fig, sample_box, space):
    triple_plot(fig, sample_box, space, npoints=[50, 100])
    fig.set_figwidth(19/2.54)

def triple_plot_50_100(fig, sample_box, space):
    triple_plot(fig, sample_box, space, npoints=[50, 100])

def triple_plot_35_50(fig, sample_box, space):
    triple_plot(fig, sample_box, space, npoints=[35, 50])
    
def triple_wide_25_50(fig, sample_box, space):
    triple_vector_plot(fig, sample_box, space, npoints=[25, 50])
    fig.set_figwidth(19/2.54)

def triple_nodes_plot(fig, sample_box, space, npoints=[25, 50]):
    ax = fig.add_subplot(131)
    ax.sample_box = sample_box[0:npoints[0]]
    ax.space = space
    try:
        tn_scheme = sample_box.Tri.tn_scheme
    except:
        import quadpy
        tn_scheme = quadpy.tn.grundmann_moeller(sample_box.nvar, 5)
    maxes.tri_nodes_plot(ax, tri_space=None, tn_scheme=tn_scheme)
    ax.set_title(space_labels[space], pad=10)
    
    ax = fig.add_subplot(132, sharex=ax, sharey=ax)
    ax.sample_box = sample_box[0:npoints[1]]
    ax.space = space
    maxes.tri_nodes_plot(ax, tri_space=None, tn_scheme=tn_scheme)
    ax.set_title(space_labels[space], pad=10)
    
    ax = fig.add_subplot(133, sharex=ax, sharey=ax)
    ax.sample_box = sample_box
    ax.space = space
    maxes.tri_nodes_plot(ax, tri_space=None, tn_scheme=tn_scheme)
    ax.set_title(space_labels[space], pad=10)

def triple_vector_plot(fig, sample_box, space, npoints=[25, 50]):
    ax = fig.add_subplot(131)
    ax.sample_box = sample_box[0:npoints[0]]
    ax.space = space
    
    maxes.tri_vector_plot(ax, tri_space=None)
    ax.set_title(space_labels[space], pad=10)
    
    ax = fig.add_subplot(132, sharex=ax, sharey=ax)
    ax.sample_box = sample_box[0:npoints[1]]
    ax.space = space
    maxes.tri_vector_plot(ax, tri_space=None)
    ax.set_title(space_labels[space], pad=10)
    
    ax = fig.add_subplot(133, sharex=ax, sharey=ax)
    ax.sample_box = sample_box
    ax.space = space
    maxes.tri_vector_plot(ax, tri_space=None)
    ax.set_title(space_labels[space], pad=10)

def triple_plot(fig, sample_box, space, npoints=[100, 200]):
    ax = fig.add_subplot(133)
    ax.sample_box = sample_box
    ax.space = space
    try:
        tn_scheme = sample_box.Tri.tn_scheme
    except:
        import quadpy
        tn_scheme = quadpy.tn.grundmann_moeller(sample_box.nvar, 5)
    maxes.convex_hull_plot(ax, **tri_kwargs[3])
    ax.xaxis.set_major_locator(ticker.FixedLocator(locs=[-6]))
    ax.yaxis.set_major_locator(ticker.FixedLocator(locs=[-6]))
    #ax.set_title(space_labels[space], pad=10)
    
    ax = fig.add_subplot(132, sharex=ax, sharey=ax)
    ax.sample_box = sample_box[0:npoints[1]]
    ax.space = space
    maxes.tri_plot(ax, **tri_kwargs[3])
    ax.xaxis.set_major_locator(ticker.FixedLocator(locs=[-6]))
    ax.yaxis.set_major_locator(ticker.FixedLocator(locs=[-6]))
    #ax.set_title(space_labels[space], pad=10)
    
    ax = fig.add_subplot(131, sharex=ax, sharey=ax)
    ax.sample_box = sample_box[0:npoints[0]]
    ax.space = space
    maxes.tri_nodes_plot(ax, tri_space=None, tn_scheme=tn_scheme, **tri_kwargs[3])
    ax.xaxis.set_major_locator(ticker.FixedLocator(locs=[-6]))
    ax.yaxis.set_major_locator(ticker.FixedLocator(locs=[-6]))
    #ax.set_title(space_labels[space], pad=10)

def qhull_under_density(fig, sample_box, space):
    from . import _axis3d_margins_patch, _axes3d
    ax3d = fig.add_subplot(111, projection='3d')
    ax3d.sample_box = sample_box
    ax3d.space = space
    maxes3d.qhull_under_density(ax3d)

def plane_under_density(fig, sample_box, space):
    from . import _axis3d_margins_patch, _axes3d
    ax3d = fig.add_subplot(111, projection='3d')
    ax3d.sample_box = sample_box
    ax3d.space = space
    maxes3d.plane_under_density(ax3d)


def dhull_vs_complete(fig, sample_box, space):
    from .. import sball
    rand_dir = sball.get_random_directions(8, 2)
    
    ax = fig.add_subplot(121)
    ax.sample_box = sample_box
    ax.space = space
    maxes.dhull_random_plot(ax, rand_dir=rand_dir)
    ax = fig.add_subplot(122)
    ax.sample_box = sample_box
    ax.space = space
    maxes.completehull_plot(ax, rand_dir=rand_dir)
    





#!/usr/bin/env python
# coding: utf-8


import numpy as np
from scipy.spatial import distance

import quadpy

from .. import shell
from .. import simplex as sx
from .. import convex_hull as khull
from ..reader import Store
from ..candynodes import CandyNodes
from .. import sball

from ._exploration import _Exploration

from collections import namedtuple
from sortedcollections import ValueSortedDict





def get_entropy(pf):
    return -pf * np.log(pf) - (1 - pf) * np.log(1 - pf)


    

CircumTriEstimation = namedtuple('CircumTriEstimation', (
    "nsim",
    "nvar",
    "nfacets",
    "r","R",
    "inner",
    "shell",
    "outer",
    "FORM_outside",
    "TwoFORM_outside",
    "orth_outside",
    "inside",
    "outside",
    "success_points",
    "failure_points",
    "success",
    "failure",
    "mix",
    "vertex_estimation",
    "weighted_vertex_estimation",
    "nsimplex",
    "tn_scheme",
    "tn_scheme_points",
    "newly_invalidated",
    "newly_estimated",
    "simplex_stats",
    "candidates_sets",
    "ncoplanar"
    ))


class CircumTri(_Exploration):
    
    #č míží nám sampling_space: Ghull umí vzorkovat outside pouze v G prostoru
    #č quadpy umístí integráční bodíky v prostoru triangulace.
    def __init__(bx, sample_box, scheme, entropy_mode='weighted', 
                    circumcenters_only=True, q=10):
        
        bx.sample_box = sample_box
        bx.scheme = scheme
        bx.entropy_mode = entropy_mode
        bx.circumcenters_only = circumcenters_only
        bx.q = q
        
        
        bx.direct_plan = quadpy.un.mysovskikh_1(bx.nvar).points
        bx.sball = sball.Sball(bx.nvar)
        
        bx.tri_space = 'G'
        
        #č přece ponechame složku pro uživatelské odhady
        #č stm kód může semka něco ukladat
        bx.estimations = []
        
        #č vítejte nové uložiště odhadů.
        #č Odhady z stm kódu už ale nemají na tohle sahat
        if hasattr(bx, 'filename'):
            bx.box_estimations = Store.create(bx.filename + "_ctri", CircumTriEstimation)
        else:
            bx.box_estimations = []
            
        bx.CC = sx.CircumCenter(sample_box.nvar)
            
        # kind of interface to CandidatesWidget
        bx.candidates_index = {}
        bx.potential_index = ValueSortedDict()
        
        bx.regen()
    
        
    def init_parameters(bx):
        """
        Returns dictionary of parameters the DiceBox was initialized with
        """
        return {'sample_box':bx.sample_box, 'scheme':bx.scheme.name, 
                'entropy_mode':bx.entropy_mode, 
                'circumcenters_only':circumcenters_only, 'q':bx.q}
                 
    
    def __repr__(bx):
        return "%s(**%s)"%(bx.__class__.__name__,  repr(bx.init_parameters()))
        
    def __str__(bx):
        return "%s(%s)"%(bx.__class__.__name__,  str(bx.init_parameters()))
        
        
        
    def refine(bx):
        if len(bx.candidates_index):
            # return node with the greatest potential
            key, value = bx.potential_index.peekitem()
            return bx.candidates_index[key]
        else:
            bx._logger(msg='refine called, but triangulation does not exist yet. Fallback to random sample')
            return bx.f_model(1)
    
    
    def regen(bx):
        """
        regen() recreates data structures of the box. 
        It shouldn't be called without reason, changed distribution, settings or so.
        """
        
        #оӵ шайтан регенираци лэзьиз
        bx._logger(msg='regeneration started')
        
        bx.candidates_index.clear()
        bx.potential_index.clear()
        
        bx._regen_outside()
        bx._regen_inside()
        
        if bx.nsim > 0:
            bx.get_pf_estimation() #č updates global stats
            bx.update_exploration_ratio()
        else:
            bx.exploration_ratio = 1
        
        bx._nsim = bx.nsim
                 
    
    
    
    
    def __len__(bx):
        return bx.sample_box.nsim
    
    def __getitem__(bx, slice):
        #č stačí vratit sample_box
        return bx.sample_box[slice]
    
    def __getattr__(dx, attr):
        if attr == 'dicebox':
            return dx
            
        #č branime sa rekurzii
        # defend against recursion
        #оӵ рекурсилы пезьдэт!
        if attr == 'sample_box':
            raise AttributeError
                
        #ё По всем вопросам обращайтесь 
        #ё на нашу горячую линию    
        else:
            return getattr(dx.sample_box, attr)
    
    
    # The DiceBox Observer 
    def _logger(self, *args, msg="", indent=0, **kwargs):
        if not kwargs:
            kwargs = "" #č ať se nám prázdné závorky nezobrazujou
        print(self.__class__.__name__ + ":", msg, *args, kwargs) 
        
        
        # inspired by Qt
    def connect(self, slot): self._logger = slot
    def disconnect(self): del(self._logger)
    
        
    # přidávání vzorků musí bejt explicitní!
    def add_sample(bx, input_sample):
        bx.sample_box.add_sample(input_sample)
        bx.increment(bx.sample_box[bx._nsim:])
        bx._nsim = bx.nsim
    
    
    
    def update_exploration_ratio(bx):
        if "tri" in dir(bx):
            outside = np.sqrt(bx.global_stats['outside'])
            mixed = np.sqrt(bx.global_stats['mix'])
            bx.exploration_ratio = outside / (outside + mixed)
        else:
            bx.exploration_ratio = 1
        
        
    
    
    
    #č bejvalej .estimate_simplex()
    #č teď je to kolbek, který volá Triangulation
    def _on_add_simplex(bx, nodes):
        # -1=outside, 0=success, 1=failure, 2=mix
        if nodes.event_id != 2:
            return 
        
        
        indices = nodes.indices
        failsi = bx.Tri.failsi[indices]
        vertices_model = bx.Tri.tri.points[indices]
        PDF = bx.Tri.PDF[indices]
        
        circum_center = bx.CC.get_circumcenter(vertices_model)
        r = distance.euclidean(circum_center, vertices_model[0])
        circum_node = bx.f_model.new_sample(circum_center, space=bx.tri_space)
        #č můžeme nechat numpy pole z jednoho prvku. Můžeme zredukovat na float
        #circum_pdf = circum_node.pdf(bx.tri_space)
        circum_pdf = float(circum_node.pdf(bx.tri_space))
        circum_potential = r * circum_pdf**(1/bx.nvar)
        
        # circum rating
        if bx.entropy_mode == 'none':
            circum_rating = circum_potential
        elif bx.entropy_mode == 'simple':
            # fp like a failure points. Number of failure points
            fp = len(failsi[failsi]) # the fastest solution
            circum_rating = circum_potential * get_entropy(fp / len(failsi))
        elif bx.entropy_mode == 'weighted':
            # same as np.average(failsi, weights=pdf), but faster
            wfr = np.sum(PDF[failsi]) / np.sum(PDF)
            circum_rating = circum_potential * get_entropy(wfr)
        
        
        if bx.circumcenters_only:
            #č nodes příjdou zabalené do CandyNodes. Ty mají .attrs a .kwargs
            circum_node = CandyNodes(circum_node, nodes.attrs)
            circum_node.potential = circum_potential
            circum_node.rating = circum_rating
            bx.candidates_index[tuple(indices)] = circum_node
            bx.potential_index[tuple(indices)] = circum_rating
            
            return
        
        #else:
        
        nodes_model = getattr(nodes, bx.tri_space)
        dr = distance.cdist(nodes_model, [circum_center]).reshape(-1)
        nodes_pdf = nodes.pdf(bx.tri_space)
        node_potentials = (r - dr) * nodes_pdf**(1/bx.nvar)
        
        if bx.entropy_mode == 'none':
            node_ratings = node_potentials
        elif bx.entropy_mode == 'simple':
            node_ratings = node_potentials * get_entropy(nodes.pfv)
        elif bx.entropy_mode == 'weighted':
            node_ratings = node_potentials * get_entropy(nodes.pfw)
            
        
        
        max_node = np.nanargmax(node_ratings)
        max_node_rating = node_ratings[max_node]
        if circum_rating > max_node_rating:
            #č nodes příjdou zabalené do CandyNodes. Ty mají .attrs a .kwargs
            circum_node = CandyNodes(circum_node, nodes.attrs)
            circum_node.potential = circum_potential
            circum_node.rating = circum_rating
            bx.candidates_index[tuple(indices)] = circum_node
            bx.potential_index[tuple(indices)] = circum_rating
        else:
            node = nodes[max_node]
            node.potential = node_potentials[max_node]
            node.rating = max_node_rating
            bx.candidates_index[tuple(indices)] = node
            bx.potential_index[tuple(indices)] = max_node_rating
            



            
    # callback
    #č sx.on_delete_simplex(indices=indices)
    def _invalidate_simplex(bx, simplex):
        bx.candidates_index.pop(simplex, None)
        bx.potential_index.pop(simplex, None)
    
    
    
        

                
    def _regen_outside(bx):
        bx.convex_hull = khull.QHull(bx.f_model, space='G', 
                                    incremental=True, auto_update=False)
        bx.ghull = Ghull(bx.convex_hull)
        #č pokud mám aspoň jednu tečku, tak už je mi šuma
        #č zda se konvexní obálka vytvořila, či nikoliv
        if bx.nsim > 0:
            bx.estimate_outside()
        
            
    def _regen_inside(bx):
        failsi = bx.failsi
        # incremental triangulation require one more point
        if (bx.nsim > bx.nvar + 1) and np.any(failsi) and not np.all(failsi):
            #bx._logger(msg="triangulation started")
            bx.__regen_inside()
        else:
            #č jíž není nutný
            bx._logger(msg="triangulation skipped")
            
    def __regen_inside(bx):
        try:
            bx.Tri = sx.BetterCubatureIntegration(bx.samplebox, bx.scheme,\
                     tri_space='G', incremental=True,\
                    on_add_simplex=bx._on_add_simplex,\
                    on_delete_simplex=bx._invalidate_simplex)
              
            bx.Tri.integrate() # nic nevrácí, všecko je přes kolbeky
            #č tri - Deloneho triangulace
            bx.tri = bx.Tri.tri #č všichní tam očekávajou QHull
            bx._logger(msg="triangulation has been created")
            
        except BaseException as e:
            #č chcu zachytit spadnuti QHull na začatku, 
            #č kdy ještě není dostatek teček.
            #č Jinak je třeba nechat QHull spadnout
            if bx.nsim > 2*bx.nvar + 3: 
                #č no to teda ne!
                raise
            else: 
                #č lze přípustit chybu triangulace    
                bx._logger(msg='triangulation failed')
    
    
        
    def increment(bx, input_sample):
        #č tri - Deloneho triangulace
        if "tri" in dir(bx):
            bx.Tri.update()
        else:
            bx._regen_inside()
            
            
        if np.any(bx.convex_hull.is_outside(input_sample)):
            bx.convex_hull.update()
            bx.estimate_outside()
        
        bx.box_estimations.append(bx.get_pf_estimation())
        bx.update_exploration_ratio()
    
                
    
        
        
    
    def estimate_outside(bx):
        #č konečně mám pořádnou stejtful třídu
        #č pokud mám aspoň jednu tečku, tak už je mi šuma
        #č zda se konvexní obálka vytvořila, či nikoliv
        
        shell_estimation, global_stats = bx.ghull.get_shell_estimation()
        
        # shell_estimation  -22: inner, -3: shell, -11: outer
        bx.shell_estimation = shell_estimation
        
        #č kvůli názvu neleze do namedtuple
        global_stats['TwoFORM_outside'] = global_stats.pop('2FORM_outside')
        #č ndim se nelibí pandas
        global_stats['nvar'] = global_stats.pop('ndim')
        
        bx.global_stats = global_stats
        
        
        
        
        
    def get_pf_estimation(bx):
        global_stats = bx.global_stats
        orth_outside = global_stats['outside'] = global_stats['orth_outside']
        pf_inside = global_stats['inside'] = 1 - orth_outside
        
        #č Ghull spouštíme sporadicky, 
        #č takže musíme sami lepit nové etikety
        global_stats['nsim'] = bx.nsim
        
        failsi = bx.failsi
        
        if 'tri' in dir(bx):
            bx.global_stats.update(bx.Tri.get_pf_estimation())
            failure = bx.global_stats['failure']
            mixed = bx.global_stats['mix']
            bx.global_stats['success_points'] = len(failsi[~failsi])
            bx.global_stats['failure_points'] = len(failsi[failsi])
            bx.global_stats['success'] = pf_inside - failure - mixed
            bx.global_stats['candidates_sets'] = len(bx.candidates_index)
            
            return CircumTriEstimation(**bx.global_stats)
            
        
        #оӵ триангуляци ӧвӧл, иськем...
        
        #č může se stát, že první dvě tečky už hned májí různé barvy,
        #č ale žádnej simplex ještě nemáme.
        #č takže celou skříňku prostě bereme jako simplex
        event, event_id, fr, wfr = sx.get_simplex_event(bx, weighting_space=bx.tri_space)
        # -1=outside, 0=success, 1=failure, 2=mix
        tri_estimation = {0:0, 1:0, 2:0}
        tri_estimation[event_id] = pf_inside
        
        vertex_estimation = pf_inside * fr
        weighted_vertex_estimation = pf_inside * wfr
        
        global_stats = bx.global_stats
        # outside dodá Ghull
        global_stats['success_points'] = len(failsi[~failsi])
        global_stats['failure_points'] = len(failsi[failsi])
        global_stats['success'] = tri_estimation[0]
        global_stats['failure'] = tri_estimation[1]
        global_stats['mix'] = tri_estimation[2]
        global_stats['vertex_estimation'] = vertex_estimation
        global_stats['weighted_vertex_estimation'] = weighted_vertex_estimation
        global_stats['nsimplex'] = 0
        global_stats['tn_scheme'] = bx.scheme.name
        global_stats['tn_scheme_points'] = bx.scheme.points.shape[1]
        global_stats['newly_invalidated'] = 0
        global_stats['newly_estimated'] = 0
        global_stats['simplex_stats'] = 0
        global_stats['candidates_sets'] = len(bx.candidates_index)
        global_stats['ncoplanar'] = 0
        
        return CircumTriEstimation(**global_stats)
        
        



#!/usr/bin/env python
# coding: utf-8


import numpy as np
from scipy.spatial import distance
from scipy.spatial import KDTree
from scipy import stats

import quadpy

from .. import shell
from .. import simplex as sx
from .. import convex_hull as khull
from ..reader import Store
from ..candynodes import CandyNodes
from .. import sball

from ._exploration import _Exploration

from collections import namedtuple
from sortedcollections import ValueSortedDict


    
    
TriEstimation = namedtuple('TriEstimation', (
    *shell.ShellStats._fields,
    *shell.ShellEstimation._fields,
    'success_points',
    'failure_points',
    *sx.CubatureEstimation._fields[2:],
    'r_safe',
    'R_safe',
    'r_mixed',
    'R_mixed',
    'r_failure',
    'R_failure',
    'p_sum',
    'max_potential'
    ))


QTriEstimation = namedtuple('QTriEstimation', (
    *khull.QHullEstimation._fields,
    'success_points',
    'failure_points',
    *sx.CubatureEstimation._fields[2:],
    'r_safe',
    'R_safe',
    'r_mixed',
    'R_mixed',
    'r_failure',
    'R_failure',
    'p_sum',
    'max_potential'
    ))


class _CircumTri(_Exploration):
    
    def __repr__(bx):
        return "%s(**%s)"%(bx.__class__.__name__,  repr(bx.init_parameters()))
        
    def __str__(bx):
        return "%s(%s)"%(bx.__class__.__name__,  str(bx.init_parameters()))
    
    def __len__(bx):
        return bx.sample_box.nsim
    
    def __getitem__(bx, slice):
        #č stačí vratit sample_box
        return bx.sample_box[slice]
    
    def __getattr__(dx, attr):
        if attr == 'dicebox':
            return dx
            
        #č branime sa rekurzii
        # defend against recursion
        #оӵ рекурсилы пезьдэт!
        if attr == 'sample_box':
            raise AttributeError
                
        #ё По всем вопросам обращайтесь 
        #ё на нашу горячую линию    
        else:
            return getattr(dx.sample_box, attr)
    
    
    # The DiceBox Observer 
    def _logger(self, *args, msg="", indent=0, **kwargs):
        if not kwargs:
            kwargs = "" #č ať se nám prázdné závorky nezobrazujou
        print(self.__class__.__name__ + ":", msg, *args, kwargs) 
        
        
        # inspired by Qt
    def connect(self, slot): self._logger = slot
    def disconnect(self): del(self._logger)
    
        
    # přidávání vzorků musí bejt explicitní!
    def add_sample(bx, input_sample):
        bx.sample_box.add_sample(input_sample)
        bx.increment(bx.sample_box[bx._nsim:])
        bx._nsim = bx.nsim
            
    
    def get_exploratory_radius(bx):
        R = bx.shell_stats.R
        # empirical rule to get desired behavior
        sphere_area = bx.convex_hull.nsphere_surface_area * (R + bx.q)**bx.nvar
        p_desired = bx.shell_estimation.outside / sphere_area
        
        # get matematically clean radius of it
        r = bx.sball.get_r(p_desired)
        return r
    
                    
    def is_mixed(bx, simplices=None):
    
        if simplices is None:
            simplices = bx.tri.simplices
        
        in_failure = np.isin(simplices, bx.failure_points)
        has_failure = in_failure.any(axis=1)
        all_failure = in_failure.all(axis=1)
        return np.logical_xor(has_failure, all_failure)
    
    
    
        
    
    def get_tri_radia(bx):
        lengths = np.sum(np.square(bx.G), axis=1)
        lengths = np.sqrt(lengths, out=lengths) #lengths of each radius-vector
        
        if 'Tri' in dir(bx):
            simplices = bx.Tri.tri.simplices
            radia = lengths[simplices]
            r = np.min(radia, axis=1)
            R = np.max(radia, axis=1)
            
            in_failure = bx.failsi[simplices]
            
            has_failure = in_failure.any(axis=1)
            all_failure = in_failure.all(axis=1)
            
            if np.any(~has_failure):
                r_safe = np.min(r[~has_failure])
                R_safe = np.max(R[~has_failure])
            else:
                r_safe, R_safe = -1, -1
            
            
            if np.any(all_failure):
                r_failure = np.min(r[all_failure])
                R_failure = np.max(R[all_failure])
            else:
                r_failure, R_failure = -1, -1
                
            mixed_mask = np.logical_xor(has_failure, all_failure)
            if np.any(mixed_mask):
                r_mixed = np.min(r[mixed_mask])
                R_mixed = np.max(R[mixed_mask]) 
            else:
                r_mixed, R_mixed = -1, -1
            
            return r_safe, R_safe, r_mixed, R_mixed, r_failure, R_failure
            
        else:
            radia = [-1] * 6
            r = np.min(lengths)
            R = np.max(lengths)
            
            if np.all(bx.failsi):
                return -1, -1, -1, -1, r, R
            elif np.any(bx.failsi):
                return -1, -1, r, R, -1, -1
            else:
                return r, R, -1, -1, -1, -1
            
            
        
    
    def perform_topological_analysis(bx):
        assert bx.tri_space == 'G'
        # tn_scheme, points, simplices, neighbors, failsi, facets, normals
        bx.ta = sx.TopologyAnalysis(bx.convex_hull.tn_scheme, bx.G, 
                                    bx.Tri.tri.simplices, bx.Tri.tri.neighbors,
                                     bx.failsi, bx.convex_hull.simplices, 
                                     bx.convex_hull.A)
        print("start integration of the convex envelope")   
        bx.ta.integrate_convex_envelope()
        print("start integration of the internal walls")
        bx.ta.integrate_walls()
        print("start topological analysis")   
        bx.ta.perform_analysis()
        
        
    
    
    
    def get_pf_estimation(bx):
        failsi = bx.failsi
        
        success_points = len(failsi[~failsi])
        failure_points = len(failsi[failsi])
        
        inside = bx.shell_estimation.inside
        
        if 'Tri' in dir(bx):
            tri_estimation = bx.Tri.get_pf_estimation()
            
            mixed = tri_estimation.mix
            failure = tri_estimation.failure
            
            if not bx.holydays:
                success = inside - mixed - failure
                p_sum = 1
            else:
                success = tri_estimation.success
                outside = bx.shell_estimation.outside
                p_sum = outside + failure + mixed + success
            
            return bx.Estimation(bx.nsim,  #č musíme sami lepit nové etikety, neboť
                                *bx.get_outside_estimation(),
                                success_points,
                                failure_points,
                                success,
                                *tri_estimation[3:],
                                
                                *bx.get_tri_radia(),
                                p_sum,
                                bx.get_max_potential()
                                )
            
        
        #оӵ триангуляци ӧвӧл, иськем...
        
        #č může se stát, že první dvě tečky už hned májí různé barvy,
        #č ale žádnej simplex ještě nemáme.
        #č takže celou skříňku prostě bereme jako simplex
        event, event_id, fr, wfr = sx.get_simplex_event(bx, weighting_space=bx.tri_space)
        # -1=outside, 0=success, 1=failure, 2=mix
        tri_estimation = {0:0, 1:0, 2:0}
        tri_estimation[event_id] = inside
        
        success = tri_estimation[0]
        failure = tri_estimation[1]
        mixed = tri_estimation[2]
        
        return bx.Estimation(bx.nsim,  #č musíme sami lepit nové etikety, neboť
                             *bx.get_outside_estimation(),
                             success_points,
                             failure_points,
                             
                             success,
                             failure,
                             mixed,
                             inside * fr,
                             inside * wfr, 
                             inside * wfr,
                             0, # nsimplex
                             #sx.tn_scheme.name,
                             bx.scheme.points.shape[1],
                             0, #sx.newly_invalidated,
                             0, #sx.newly_estimated,
                             0, #len(sx.failure_simplices),
                             0, #len(sx.mixed_simplices),
                             0, #len(sx.tri.coplanar),
                             
                             *bx.get_tri_radia(),
                             1,
                             bx.get_max_potential()
                             )
        







class CirQTri(_CircumTri):
    
    def __init__(bx, sample_box, scheme, convex_hull_degree=5,
                 q=1, holydays=0):
        
        bx.sample_box = sample_box
        bx.convex_hull_degree = convex_hull_degree
        bx.scheme = scheme
        
        bx.tri_space = 'G'
            
        bx.Estimation = QTriEstimation
        bx.q = q
        bx.holydays = holydays
        
        bx.sball = sball.Sball(bx.nvar)
        
        
        #č přece ponechame složku pro uživatelské odhady
        #č stm kód může semka něco ukladat
        bx.estimations = []
        
        #č vítejte nové uložiště odhadů.
        #č Odhady z stm kódu už ale nemají na tohle sahat
        if hasattr(bx, 'filename'):
            bx.box_estimations = Store.create(bx.filename + "_qtri", bx.Estimation)
        else:
            bx.box_estimations = []
            
        bx.CC = sx.CircumCenter(sample_box.nvar)
        
        bx.regen()
    
        
    def init_parameters(bx):
        """
        Returns dictionary of parameters the DiceBox was initialized with
        """
        return {'sample_box':bx.sample_box, 'scheme':bx.scheme.name,
                'convex_hull_degree':bx.convex_hull_degree,
                 'q':bx.q, 'holydays':bx.holydays}


    def _regen_outside(bx):
        facet_scheme = quadpy.tn.grundmann_moeller(bx.nvar - 1, bx.convex_hull_degree)
        bx.convex_hull = khull.QHullCubature(bx.f_model, space='G', incremental=True, 
                            auto_update=False, tn_scheme=facet_scheme)
        
        #č konečně mám pořádnou stejtful třídu
        #č pokud mám aspoň jednu tečku, tak už je mi šuma
        #č zda se konvexní obálka vytvořila, či nikoliv
        if bx.nsim > 0:
            bx.estimate_outside()
            
    
    def estimate_outside(bx):
        #č konečně mám pořádnou stejtful třídu
        #č pokud mám aspoň jednu tečku, tak už je mi šuma
        #č zda se konvexní obálka vytvořila, či nikoliv
        
        #č Máme úkol: 
        #č 1. Získat odhady a uložit je, abychom nemuseli opakovaně integrovat,
        #č    dokud se neobjeví nějaký nový vzorek zvenku.
        bx.shell_estimation = bx.convex_hull.get_convex_hull_estimation()
        bx.shell_stats = bx.shell_estimation # simple trick
    
            
    def get_outside_estimation(bx):
        return bx.shell_estimation[1:]
    
    
    def _regen_inside(bx):
        failsi = bx.failsi
        # incremental triangulation require one more point
        if (bx.nsim > bx.nvar + 1) and np.any(failsi) and not np.all(failsi):
            try:
                bx.Tri = sx.GaussCubatureIntegration(bx.samplebox, bx.scheme, 
                                            incremental=True, full=bx.holydays)
                
                #č tri - Deloneho triangulace
                bx.tri = bx.Tri.tri #č všichni tam očekávají QHull
                bx._logger(msg="triangulation has been created")
                
            except BaseException as e:
                #č chcu zachytit spadnuti QHull na začatku, 
                #č kdy ještě není dostatek teček.
                #č Jinak je třeba nechat QHull spadnout
                if bx.nsim > 2*bx.nvar + 3: 
                    #č no to teda ne!
                    raise
                else: 
                    #č lze přípustit chybu triangulace    
                    bx._logger(msg='triangulation failed')
        else:
            #č jíž není nutný
            bx._logger(msg="triangulation skipped")
            
    
    def get_circum_node(bx, indices):
        vertices_model = bx.Tri.tri.points[indices]
        try:
            circum_center = bx.CC.get_circumcenter(vertices_model)
        except BaseException as e:
            bx._logger(repr(e), msg='CircumCenter error')
            circum_center = np.mean(vertices_model, axis=0)
             
        if not np.all(np.isfinite(circum_center)):
            raise
             
        length2 = np.sum(np.square(circum_center))
        R = bx.get_exploratory_radius()
        if length2 > R**2:
            bx._logger(circum_center, 
                    msg='Circumcenter lies too far away from the origin. The radial distance is reduced')
            circum_center /= np.sqrt(length2)
            circum_center *= R
        
        circum_node = bx.f_model.new_sample(circum_center, space=bx.tri_space)
        return circum_node
    
    
    def refine(bx):
        if "Tri" in dir(bx):
            # get the greatest mixed simplex
            return bx.get_circum_node(bx.Tri.max_mixed_indices)
        else:
            bx._logger(msg='refine called, but triangulation does not exist yet. Fallback to a random sample')
            return bx.f_model(1)
    
    
    def holyday(bx):
        if "Tri" in dir(bx) and (bx.Tri.max_safe > 0):
            return bx.get_circum_node(bx.Tri.max_safe_indices)
        else:
            bx._logger(msg='Fallback to random sample')
            return bx.f_model(1)
    
    def regen(bx):
        """
        regen() recreates data structures of the box. 
        It shouldn't be called without reason, changed distribution, settings or so.
        """
        
        #оӵ шайтан регенираци лэзьиз
        bx._logger(msg='regeneration started')
        
        bx._regen_outside()
        bx._regen_inside()
        
        bx.to_explore = 1
        bx.to_refine = 0
        if bx.nsim > 0:
            bx.pf_estimation = bx.get_pf_estimation() 
            bx.update_exploration_ratio()
        
        bx._nsim = bx.nsim



    def get_max_potential(bx):
        if "Tri" in dir(bx):
             # get probability of the greatest mixed simplex
            return bx.Tri.max_mixed
        else:
            return -1

    def update_exploration_ratio(bx):
        bx.to_explore = stats.norm.sf(bx.pf_estimation.r)
        bx.to_refine = bx.get_max_potential()

        
    def increment(bx, input_sample):
        #č tri - Deloneho triangulace
        if "tri" in dir(bx):
            bx.Tri.update()
        else:
            bx._regen_inside()
            
        if np.any(bx.convex_hull.is_outside(input_sample)):
            bx.convex_hull.update()
            bx.estimate_outside()
        
        bx.pf_estimation = bx.get_pf_estimation()
        bx.box_estimations.append(bx.pf_estimation)
        bx.update_exploration_ratio()




class _PsiTri:
    def refine(bx):
        if len(bx.candidates_index):
            # return node with the greatest potential
            key, value = bx.potential_index.peekitem()
            return bx.candidates_index[key]
        else:
            bx._logger(msg='refine called, but triangulation does not exist yet. Fallback to random sample')
            return bx.f_model(1)
    
    
    def holyday(bx):
        if len(bx.candidates_index):
            # return node with the greatest potential
            key, value = bx.global_potential_index.peekitem()
            return bx.candidates_index[key]
        else:
            bx._logger(msg='triangulation does not exist yet. Fallback to random sample')
            return bx.f_model(1)
    
    def regen(bx):
        """
        regen() recreates data structures of the box. 
        It shouldn't be called without reason, changed distribution, settings or so.
        """
        
        #оӵ шайтан регенираци лэзьиз
        bx._logger(msg='regeneration started')
        
        bx.candidates_index.clear()
        bx.potential_index.clear()
        bx.global_potential_index.clear()
        
        bx._regen_outside()
        bx._regen_inside()
        
        bx.to_explore = 1
        bx.to_refine = 0
        if bx.nsim > 0:
            bx.pf_estimation = bx.get_pf_estimation() #č updates p_mixed
            bx.update_exploration_ratio()
        
        bx._nsim = bx.nsim


    def _regen_inside(bx):
        failsi = bx.failsi
        # incremental triangulation require one more point
        if (bx.nsim > bx.nvar + 1) and np.any(failsi) and not np.all(failsi):
            #bx._logger(msg="triangulation started")
            bx.__regen_inside()
        else:
            #č jíž není nutný
            bx._logger(msg="triangulation skipped")
            
    def __regen_inside(bx):
        try:
            if bx.holydays:
                Integrator = sx.FullCubatureIntegration
            else:
                Integrator = sx.FastCubatureIntegration
            
            bx.Tri = Integrator(bx.samplebox, bx.scheme,\
                     tri_space=bx.tri_space, incremental=True,\
                    on_safe_added=bx._on_safe_added,\
                    on_mixed_added=bx._on_mixed_added,\
                    on_delete_simplex=bx._invalidate_simplex)
              
            bx.Tri.integrate() # nic nevrácí, všecko je přes kolbeky
            #č tri - Deloneho triangulace
            bx.tri = bx.Tri.tri #č všichní tam očekávajou QHull
            bx._logger(msg="triangulation has been created")
            
        except BaseException as e:
            #č chcu zachytit spadnuti QHull na začatku, 
            #č kdy ještě není dostatek teček.
            #č Jinak je třeba nechat QHull spadnout
            if bx.nsim > 2*bx.nvar + 3: 
                #č no to teda ne!
                raise
            else: 
                #č lze přípustit chybu triangulace    
                bx._logger(msg='triangulation failed')

    def get_max_potential(bx):
        if len(bx.candidates_index):
             __key, max_potential = bx.potential_index.peekitem()
        else:
            max_potential = -1
        return max_potential



    def update_exploration_ratio(bx):
        bx.to_explore += bx.pf_estimation.outside / (bx.pf_estimation.nfacets + 1)
        if bx.pf_estimation.max_potential > 0:
            bx.to_refine += bx.pf_estimation.mix / bx.pf_estimation.mixed_simplices


    #č teď je to kolbek, který volá Triangulation
    def _on_safe_added(bx, simplex, indices, vertices_model, nodes, vol):
        
        try:
            circum_center = bx.CC.get_circumcenter(vertices_model)
        except BaseException as e:
            bx._logger(repr(e), msg='CircumCenter error')
            return 
        r = distance.euclidean(circum_center, vertices_model[0])
        circum_node = bx.f_model.new_sample(circum_center, space=bx.tri_space)
        #č můžeme nechat numpy pole z jednoho prvku. Můžeme zredukovat na float
        #circum_pdf = circum_node.pdf(bx.tri_space)
        circum_pdf = float(circum_node.pdf(bx.tri_space))
        
        
        #PDF = bx.Tri.PDF[indices]
        #max_PDF = np.max(PDF)
        
        circum_potential = r**bx.nvar * circum_pdf #**bx.psi_q * max_PDF**(1-bx.psi_q)
        
        if not np.isfinite(circum_potential):
            return
        
#        max_node_potential = -np.inf
#        if not bx.circumcenters_only:
#            dr = distance.cdist(nodes.G, [circum_center]).reshape(-1)
#            nodes_pdf = nodes.pdf(bx.tri_space)
#            node_potentials = (r - dr)**bx.nvar * nodes_pdf#**bx.psi_q * PDF[i]**(1-bx.psi_q)
#            
#            max_node = np.argmax(node_potentials)
#            max_node_potential = node_potentials[max_node]
#        
#        if max_node_potential > circum_potential:
#            result_node = nodes[max_node]
#            result_potential = max_node_potential
#        else:
        result_node = circum_node
        result_potential = circum_potential
            
        if bx.store_candidates_metainformation:
            result_node = CandyNodes(result_node)
            result_node.potential = result_potential
            
        bx.candidates_index[simplex] = result_node
        bx.global_potential_index[simplex] = result_potential
    
    
    
    #č teď je to kolbek, který volá Triangulation
    def _on_mixed_added(bx, simplex, indices, vertices_model, nodes, _vol, fr, wfr, _mfpdf):
        
        try:
            circum_center = bx.CC.get_circumcenter(vertices_model)
        except BaseException as e:
            bx._logger(repr(e), msg='CircumCenter error')
            return 
        r = distance.euclidean(circum_center, vertices_model[0])
        circum_node = bx.f_model.new_sample(circum_center, space=bx.tri_space)
        #č můžeme nechat numpy pole z jednoho prvku. Můžeme zredukovat na float
        #circum_pdf = circum_node.pdf(bx.tri_space)
        circum_pdf = float(circum_node.pdf(bx.tri_space))
        
        
        failsi = bx.Tri.failsi[indices]
        PDF = bx.Tri.PDF[indices]
        max_PDF = np.max(PDF)
        
        circum_potential = r**bx.nvar * circum_pdf**bx.psi_q * max_PDF**(1-bx.psi_q)
        
        if not np.isfinite(circum_potential):
            return
        
        
        
        max_node_potential = -np.inf
        if not bx.circumcenters_only:
            tree = KDTree(vertices_model, compact_nodes=False, balanced_tree=False)
            
            #č konkretně tato třída je pevně napojena na G prostor
            #č ale bacha, kbyby se to změnilo...
            dd, ii = tree.query(nodes.G, k=2)
            
            failsi_pair = failsi[ii]
            mask = np.logical_xor(failsi_pair[:,0], failsi_pair[:,1])
            #mask = np.sum(failsi[ii], axis=1) == 1
            if np.any(mask):
                nodes = nodes[mask]
                
                dr = distance.cdist(nodes.G, [circum_center]).reshape(-1)
                nodes_pdf = nodes.pdf(bx.tri_space)
                #č nejdřív rouška, pak slice. V tomto pořadí numpy to zvladné bez alokací
                i = ii[mask][:,0] 
                node_potentials = (r - dr)**bx.nvar * nodes_pdf**bx.psi_q * PDF[i]**(1-bx.psi_q)
                
                max_node = np.argmax(node_potentials)
                max_node_potential = node_potentials[max_node]
        
        
        if max_node_potential > circum_potential:
            result_node = nodes[max_node]
            result_potential = max_node_potential
        else:
            result_node = circum_node
            result_potential = circum_potential
            
        if bx.store_candidates_metainformation:
            result_node = CandyNodes(result_node)
            result_node.potential = result_potential
            
        bx.candidates_index[simplex] = result_node
        bx.potential_index[simplex] = result_potential
        if bx.holydays:
            bx.global_potential_index[simplex] = result_potential
            
        
            

            
    # callback
    #č sx.on_delete_simplex(indices=indices)
    def _invalidate_simplex(bx, simplex):
        bx.potential_index.pop(simplex, None)
        bx.global_potential_index.pop(simplex, None)
        bx.candidates_index.pop(simplex, None)
        
        
    def increment(bx, input_sample):
        #č tri - Deloneho triangulace
        if "tri" in dir(bx):
            bx.Tri.update()
        else:
            bx._regen_inside()
            
        a, b, _sample = bx.convex_hull.get_exploration_vector()
        critical_r = -b
        
        coming_r = a @ input_sample.G.T  
        
        if bx._nsim > 1 and np.any(coming_r > critical_r):
            bx.to_explore = 0
            bx.convex_hull.update()
            bx.estimate_outside()
        elif np.any(bx.convex_hull.is_outside(input_sample)):
            bx.to_refine = 0
            bx.convex_hull.update()
            bx.estimate_outside()
        else:
            bx.to_refine = 0
        
        bx.pf_estimation = bx.get_pf_estimation()
        bx.box_estimations.append(bx.pf_estimation)
        bx.update_exploration_ratio()
        
        
        
        



class CircumTri(_CircumTri, _PsiTri):
    
    #č míží nám sampling_space: Ghull umí vzorkovat outside pouze v G prostoru
    #č quadpy umístí integráční bodíky v prostoru triangulace.
    def __init__(bx, sample_box, scheme, q=10, psi_q=0.5, circumcenters_only=True, 
                  holydays=10, store_candidates_metainformation=False, shell_budget=1000):
        
        bx.sample_box = sample_box
        bx.scheme = scheme
        
        bx.tri_space = 'G'
            
        bx.Estimation = TriEstimation
        bx.shell_budget = shell_budget
        #bx.potential_mode = potential_mode
        #bx.weighted_entropy = weighted_entropy
        bx.store_candidates_metainformation = store_candidates_metainformation
        bx.circumcenters_only = circumcenters_only
        bx.q = q
        bx.psi_q = psi_q
        bx.holydays = holydays
        
        bx.direct_plan = quadpy.un.mysovskikh_1(bx.nvar).points
        bx.sball = sball.Sball(bx.nvar)
        
        
        #č přece ponechame složku pro uživatelské odhady
        #č stm kód může semka něco ukladat
        bx.estimations = []
        
        #č vítejte nové uložiště odhadů.
        #č Odhady z stm kódu už ale nemají na tohle sahat
        if hasattr(bx, 'filename'):
            bx.box_estimations = Store.create(bx.filename + "_tri", bx.Estimation)
        else:
            bx.box_estimations = []
            
        bx.CC = sx.CircumCenter(sample_box.nvar)
            
        # kind of interface to CandidatesWidget
        bx.candidates_index = {}
        bx.potential_index = ValueSortedDict()
        bx.global_potential_index = ValueSortedDict()
        
        bx.regen()
    
        
    def init_parameters(bx):
        """
        Returns dictionary of parameters the DiceBox was initialized with
        """
        return {'sample_box':bx.sample_box, 'scheme':bx.scheme.name,
                 'shell_budget':bx.shell_budget,
                 'circumcenters_only':bx.circumcenters_only, 
                 'q':bx.q, 'psi_q':bx.psi_q, 'holydays':bx.holydays}


    def _regen_outside(bx):
        bx.convex_hull = khull.QHull(bx.f_model, space='G', 
                                    incremental=True, auto_update=False)
        bx.ghull = shell.GaussianAnnulus(bx.convex_hull)
        #č konečně mám pořádnou stejtful třídu
        #č pokud mám aspoň jednu tečku, tak už je mi šuma
        #č zda se konvexní obálka vytvořila, či nikoliv
        if bx.nsim > 0:
            bx.estimate_outside()
            
    
    def _rate_outside_nodes(bx, outside_nodes, d, i):
        failsi = bx.failsi
        if np.all(failsi) or not np.any(failsi):
            return
        
        try:
            mask = bx.is_mixed(bx.convex_hull.simplices[i])
            if not np.any(mask):
                return
            outside_nodes = outside_nodes[mask]
            d = d[mask]
            
        except:
            pass
            
        #č konkretně tato třída je pevně napojena na G prostor
        #č ale bacha, kbyby se to změnilo...
        _d1, i1 = bx.tree.query(outside_nodes.G, k=1)
        
        PDF = bx.pdf(bx.tri_space)
        
        nodes_pdf = outside_nodes.pdf(bx.tri_space)
        node_potentials = d**bx.nvar * nodes_pdf**bx.psi_q * PDF[i1]**(1-bx.psi_q)
        
        max_node = np.argmax(node_potentials)
        max_node_potential = node_potentials[max_node]
        
        
        if max_node_potential > bx.potential_index[-1]:
            if bx.store_candidates_metainformation:
                outside_node = CandyNodes(outside_nodes[max_node])
                outside_node.potential = max_node_potential
            else:
                outside_node = outside_nodes[max_node]
            
            bx.candidates_index[-1] = outside_node
            bx.potential_index.pop(-1)
            bx.potential_index[-1] = max_node_potential
            bx.global_potential_index.pop(-1)
            bx.global_potential_index[-1] = max_node_potential
    
    
    def estimate_outside(bx):
        #č konečně mám pořádnou stejtful třídu
        #č pokud mám aspoň jednu tečku, tak už je mi šuma
        #č zda se konvexní obálka vytvořila, či nikoliv
        
        #č Máme 2 úkoly: 
        #č 1. Získat odhady a uložit je, abychom nemuseli opakovaně integrovat,
        #č    dokud se neobjeví nějaký nový vzorek zvenku.
        #č 2. Získat kandidaty.
        #č    a. z mezíkruží (-12)
        
        #č prace s tečkami v mezikruži se změnila
        #č teď tečky dostávám přes kolbek po částech
        #č a není předem známo, kolik těch částí bude.
        bx.candidates_index.pop(-1, None)
        #č těch kastomných slovníků se bojím...
        bx.potential_index.pop(-1, None)
        bx.global_potential_index.pop(-1, None)
        #bx.rating_index.pop(-1, None)
        bx.potential_index[-1] = 0 #č pro callback
        bx.global_potential_index[-1] = 0 #č pro callback
        
        bx.shell_stats = bx.ghull.get_shell_estimation()
        
        bx.tree = KDTree(bx.G, compact_nodes=True, balanced_tree=False)
        
        # get candidates!
        #č explicitně (pokažde) počtem teček zadavám přesnost integrace
        #č takže změny bx.shell_budget budou při dálším spuštění aplikovány
        bx.shell_estimation = bx.ghull.integrate(bx.shell_budget,
                                callback_outside=bx._rate_outside_nodes)         
    
            
    def get_outside_estimation(bx):
        #č Ghull spouštíme sporadicky
        return *bx.shell_stats[1:], *bx.shell_estimation
            
            

class QTri(_CircumTri, _PsiTri):
    
    def __init__(bx, sample_box, convex_hull_degree=5, tri_degree=5,
                 q=10, psi_q=0.5, circumcenters_only=True, 
                  holydays=10, store_candidates_metainformation=False):
        
        bx.sample_box = sample_box
        bx.convex_hull_degree = convex_hull_degree
        bx.tri_degree = tri_degree
        bx.scheme = quadpy.tn.grundmann_moeller(bx.nvar, tri_degree)
        
        bx.tri_space = 'G'
            
        bx.Estimation = QTriEstimation
        bx.store_candidates_metainformation = store_candidates_metainformation
        bx.circumcenters_only = circumcenters_only
        bx.q = q
        bx.psi_q = psi_q
        bx.holydays = holydays
        
        bx.sball = sball.Sball(bx.nvar)
        
        
        #č přece ponechame složku pro uživatelské odhady
        #č stm kód může semka něco ukladat
        bx.estimations = []
        
        #č vítejte nové uložiště odhadů.
        #č Odhady z stm kódu už ale nemají na tohle sahat
        if hasattr(bx, 'filename'):
            bx.box_estimations = Store.create(bx.filename + "_qtri", bx.Estimation)
        else:
            bx.box_estimations = []
            
        bx.CC = sx.CircumCenter(sample_box.nvar)
            
        # kind of interface to CandidatesWidget
        bx.candidates_index = {}
        bx.potential_index = ValueSortedDict()
        bx.global_potential_index = ValueSortedDict()
        
        bx.regen()
    
        
    def init_parameters(bx):
        """
        Returns dictionary of parameters the DiceBox was initialized with
        """
        return {'sample_box':bx.sample_box, 
                'convex_hull_degree':bx.convex_hull_degree,
                'tri_degree':bx.tri_degree,
                 'circumcenters_only':bx.circumcenters_only, 
                 'q':bx.q, 'psi_q':bx.psi_q, 'holydays':bx.holydays}


    def _regen_outside(bx):
        facet_scheme = quadpy.tn.grundmann_moeller(bx.nvar - 1, bx.convex_hull_degree)
        bx.convex_hull = khull.QHullCubature(bx.f_model, space='G', incremental=True, 
                            auto_update=False, tn_scheme=facet_scheme)
        
        #č konečně mám pořádnou stejtful třídu
        #č pokud mám aspoň jednu tečku, tak už je mi šuma
        #č zda se konvexní obálka vytvořila, či nikoliv
        if bx.nsim > 0:
            bx.estimate_outside()
            
    
    def estimate_outside(bx):
        #č konečně mám pořádnou stejtful třídu
        #č pokud mám aspoň jednu tečku, tak už je mi šuma
        #č zda se konvexní obálka vytvořila, či nikoliv
        
        #č Máme úkol: 
        #č 1. Získat odhady a uložit je, abychom nemuseli opakovaně integrovat,
        #č    dokud se neobjeví nějaký nový vzorek zvenku.
        bx.shell_estimation = bx.convex_hull.get_convex_hull_estimation()
        bx.shell_stats = bx.shell_estimation # simple trick
    
            
    def get_outside_estimation(bx):
        return bx.shell_estimation[1:]
        
        



#!/usr/bin/env python
# coding: utf-8

"""
Zde leží whiteboxy
Zatimco BlackBox pěčlivě ukladá věškerá data,
věškeré sady vzorků, průběžné odhady a tak,
WhiteBox musí obsahovat jen pf_exact, které buď už předem zná,
nebo jej spočítá a není vůbec/nachren nutný cokoliv ukladat.

en:
f_model + g_model = pf
WhiteBox = [f_model, g_model, pf_exact]
pf_exact is most important part of WhiteBox 
Knowledge of pf is the only reason to create WhiteBox

whitebox actually IS g_model PLUS:
.f f_model
.pf_exact
.pf_exact_method
"""
import numpy as np
from scipy import stats
from scipy import special # for S_ball # for Z_prod
from scipy import integrate # for S_ball

import mpmath # for Gaussian_Z_min

mpmath.mp.dps = 325 # to cover anything that double-presigion float can handle

from .samplebox import SampleBox

from . import f_models
import copy
from . import g_models
from . import misc as wmisc



class WhiteBox:
    """
    Bazová třida pro dědictví
    
    úkolem whiteboxu je spočítat pf-ko
    .pf_exact
    .pf_exact_method
    """
    pf_exact_method = 'None'
    
    def __init__(wt,  f_model, g_model):
        wt.gm = g_model
        wt.f = f_model
        # na začatku nemáme vzorky - pouze rozdělení a podpís
        wt.sample_box = SampleBox(wt.f(), gm_signature=wt.gm_signature)
        
        try: # no to jsme líný. Možná samotný g_model tuší jak se spočte pf-ko?
            wt.pf_exact, wt.pf_exact_method = g_model.pf_expression(f_model)
        except AttributeError:
            pass
        
    def __str__(wt):
        #ёč хощется přidat nějaký popísek k testkejsům
        #č chce se odlišovat úlohy, co už získaly slavu 
        #č v časopísech od těch, co ještě nie,
        #č od těch, se kterými ještě hrajeme
        #č od těch, co jsme před chvilkou vymysleli
        if 'description' in wt.__dict__:
            return  wt.description
        else:
            return  wt.__class__.__name__ + ' of' + str(wt.gm)
        
    def __repr__(wt):
        return  'WhiteBox(%s, %s)' %(repr(wt.f), repr(wt.gm))
        
    def __len__(wt):
        return len(wt.sample_box)
        
        
    def __call__(wt, sample=None):
        if sample is None:
            sample = wt.sample_box()
        
        # zamykame se do sebe
        return wt.run_sample(sample)
    
    def run_sample(wt, sample):
        result = wt.gm(sample)
        wt.sample_box.add_sample(result)
        return result
        
    def __getitem__(self, slice):
        self_copy = copy.copy(self)
        self_copy.sample_box = self.sample_box[slice]
        return self_copy
        
    def __getattr__(wt, attr):
        if attr == 'whitebox':
            return wt
        # co patři g_modelovi?
        elif attr == 'get_2D_R_boundary':
            return wt.gm.get_2D_R_boundary
        elif attr == 'gm_signature':
            try: # byla volná funkce?
                return wt.gm.__name__
                # asi to byla trida?
            except AttributeError:
                return repr(wt.gm)
        
        
        # co mělo být definováno ve WhiteBoxu? Ale teda není?
        elif attr == 'pf_exact_method':
            raise AttributeError
            
          # пытка-непытка
        elif attr == 'pf_exact': 
            if 'beta_exact' in wt.__dict__:
                return stats.norm.cdf(-wt.beta_exact)
            else:
                # мы нищего не знаем
                raise AttributeError("Nothing known about failure probability")
                
        elif attr == 'beta_exact':
            if 'pf_exact' in wt.__dict__:
                return -stats.norm.ppf(wt.pf_exact)
            else:
                raise AttributeError("Nothing known about failure probability")
        
        
        # branime sa rekurzii
        # defend against recursion
        elif attr == 'sample_box':
            raise AttributeError
            
        # zbytek teda nievím
        else:
            return getattr(wt.sample_box, attr)
    
    # just plot, green points, red points...
    # commented out in order to not import 
    # matplotlib and pyqtgraph unnecessarly
    #plot2D = plot.plot2D
    #plot3D = plot.plot3D
    #show2D = plot.show2D
    #show3D = plot.show3D
            
        
    def get_2D_boundary(wt, nrod=100, viewport_sample=None, viewport_space='R'):
        """
        Fence off!
        nrod - number of rods in fencing
        viewport_sample - limit points of viewport
        (function will get xlim and ylim from there, 
        assuming there will be at least two points)
        """
        
        if (viewport_sample is not None): # and viewport_sample.nsim: #>0
            if viewport_space=='R':
                viewport_sample_R = viewport_sample
            else:
                viewport_sample_R = wt.f_model.new_sample(viewport_sample, viewport_space, extend=True).R
        else:
            viewport_sample_R = wt.f_model.new_sample([[-7,-7],[7,7]], 'G', extend=True).R
            
            # should I tolerate nD?
        viewport_sample_R = np.vstack((viewport_sample_R[:,:2], wt.sample_box.R[:,:2]))
        xmin = np.ma.masked_invalid(viewport_sample_R[:,0]).min()
        xmax = np.ma.masked_invalid(viewport_sample_R[:,0]).max()
        ymin = np.ma.masked_invalid(viewport_sample_R[:,1]).min()
        ymax = np.ma.masked_invalid(viewport_sample_R[:,1]).max()
        
        # získám seznam polí
        bounds_R = wt.get_2D_R_boundary(nrod, (xmin, xmax), (ymin, ymax))
        # transformuji na seznam vzorků
        return [wt.f_model.new_sample(bounds_R[i], extend=True)  for i in range(len(bounds_R))]
    
    
    def get_sensitivities(wt, nrod=200, G_radius=7, space='R'):
        #č nejdřív musíme získat relevantní ve skutečném prostoru rámec
        view_sample = wmisc.get_isodistances(wt.f_model, G_radius, nrod)
        view_sample_R = view_sample.R
        
        #č jakože tímto vyfiltrujeme nejen NaN, ale i nekonečna?
        xmin = np.ma.masked_invalid(view_sample_R[:,0]).min()
        xmax = np.ma.masked_invalid(view_sample_R[:,0]).max()
        ymin = np.ma.masked_invalid(view_sample_R[:,1]).min()
        ymax = np.ma.masked_invalid(view_sample_R[:,1]).max()
        
        #č skoro hotovo. Jsme připravení získat seznam hraničních bodů od g_modelu
        boundaries = wt.get_2D_R_boundary(nrod, (xmin, xmax), (ymin, ymax))
        
        sensitivity = np.zeros(2)
        # boundaries are instances of an obscure Ingot class.
        # Relicts from the WellMet's childhood
        for boundary in boundaries:
            #č podle ono to tam musí implicitně vzít podle R
            sample = wt.f_model.new_sample(boundary, extend=True)
            sample_space = getattr(sample, space)
            vectors = np.diff(sample_space, axis=0)
            #č k těm hustotám můžeme podstupovat několika způsoby
            #č buď zprůměrovat uzly, v nich napočítat hustoty.
            #č Nebo napočítat hustoty v tam, kde je máme, 
            #č zprůměrovat už jen hustoty. 
            #č Jdeme na tu druhou variantu
            pdf = sample.pdf(space)
            pdf[:-1] += pdf[1:]
            #pdf /= 2
            pdf = pdf[:-1]
            
            #č pokud nebudeme chtit nic křeslit, 
            #č tak samotné vektory ani nepotřebujeme
            np.square(vectors, out=vectors)
            
            #č zdá se, že ani délky nepotřebujeme
            #lengths = np.sum(vectors, axis=1)
            #vectors /= lengths
            
            vectors *= pdf.reshape(-1,1)
            mask = np.all(np.isfinite(vectors), axis=1)
            sensitivity += np.sum(vectors[mask], axis=0)[:2]
            
        sensitivity /= np.sum(sensitivity)
        y, x = sensitivity
        return x, y
        


    # Monte Carlo, n-krátá realizace
    def MC(wt, Nsim=int(1e6)):
        
        
        # tohlensto může bejt dost těžkým
        result = wt.gm(wt.f(Nsim))
        # should I stay or should I go?
        #wt.sample_box.add_sample(result)
        
        # je tu jakoby že g_model vždy vrací nějakej sample_box
        pf_exact = np.count_nonzero(result.failsi)/Nsim
        
        # šlo by to?
        if wt.pf_exact_method == 'None' or wt.pf_exact_method == 'MC' and wt.Nsim <= Nsim:
            wt.pf_exact = pf_exact
            wt.Nsim = Nsim
            wt.pf_exact_method = 'MC'
        
        print('Monte Carlo estimation of pf is %s (%s simulations)'%(pf_exact, Nsim))
        return result
        


    # IS, (n-2)-krátá realizace, n>2
    def IS(wt, Nsim=int(1e4), h_model=None, IS_mode='G'):
        """
        IS_mode - v jakých souřadnicích robím merge a jaká PDF použiváme?
        může být 'R' nebo 'G'
        jinde # čo jinde?
        """
        
        if h_model is not None:
            wt.h = h_model 
            wt.IS_mode = IS_mode
        elif 'h' not in wt.__dict__:
            wt.h = f_models.UnCorD([stats.norm(0,2.5) for i in range(wt.f.nvar)])
            wt.IS_mode = 'G'
            
            
        #
        # jdeme na to, koťě!
        #
        
        # zgenerujeme vzorky
        # nic zajimavýho
        wt.h = wt.h(Nsim)
        
        # a teď bacha!
        if wt.IS_mode == 'R':
            # jestli máme to právé vzorkovácí rozdělení - tak nemáme čo robiť
            to_sample = wt.f.new_sample(wt.h) # smerdží se to po R
            # w like weights
            wt.w = to_sample.pdf_R / wt.h.pdf_R
        elif wt.IS_mode == 'G':
            # tady musíme provést jeden trik
            to_sample = wt.f.new_sample(wt.h.R, 'G') # R-ko smerdžíme ako G-čko
            wt.w = to_sample.pdf_G / wt.h.pdf_R # snad je to správně
        else:
            # шо-то тут не то...
            # čo blbnéš, kámo?
            # What's going on with my IS_mode?
            raise ValueError("IS_mode should be either 'R' or 'G'")
        
        # vahy máme, jedeme dál
        # sample_box jíž není prázdnej
        result = wt.gm(to_sample)
        #wt.sample_box.add_sample(result)
        
        # hodilo by sa to?
        pf_exact = np.sum(wt.w[result.failsi])/Nsim
        
        if wt.pf_exact_method in ('None', 'IS_norm', 'IS'):
            wt.Nsim = Nsim
            if pf_exact < 1:
                wt.pf_exact = pf_exact
                wt.pf_exact_method = 'IS'
            else:
                # ať mně nerobí ostudu
                wt.pf_exact = np.sum(wt.w[result.failsi]) / np.sum(wt.w)
                wt.pf_exact_method = 'IS_norm'
        
        print('Importance Sampling pure estimation of pf is %s (%s simulations)'%(pf_exact, Nsim))
        return result
       


class HyperPlane(WhiteBox): # куда ж без него...
    def __init__(self, betas=(1,2,3)):
        """
        Class takes for inicialization tuple of betas 
        Betas are coeffitients in sense of Regression Analysis (well, not really)
        g= a*X1 + b*X2 + c
        betas=(a,b,c)
        """
        self._betas = betas
        self.gm = g_models.Linear_nD(betas)
        self.f = f_models.SNorm(len(betas)-1)
        # na začatku nemáme vzorky - pouze rozdělení a podpís
        self.sample_box = SampleBox(self.f(), gm_signature=self.gm_signature)
        
        # tady už je to ta, "náše" beta )
        # beta = c/np.sqrt(a**2 + b**2)
        self.beta_exact = betas[-1]/np.sqrt(np.sum(np.array(betas[:-1])**2)) 
        self.pf_exact = stats.norm.cdf(-self.beta_exact)
        self.pf_exact_method = 'FORM (exact solution)' # Ang, Tang and Pythagoras
        self.r_exact = self.beta_exact
        
        
    def __str__(wt):
        return  'HyperPlaneBox%sD'%(len(wt._betas)-1)
        
    def __repr__(wt):
        return  'HyperPlane(%s)' % repr(wt._betas)
        
        
class Line(WhiteBox):
    def __init__(self, nvar=2, beta=5):
        """
        Class takes for inicialization tuple of betas 
        Betas are coeffitients in sense of Regression Analysis (well, not really)
        g= a*X1 + b*X2 + c
        betas=(a,b,c)
        """
        self.beta_exact = beta
        self.gm = g_models.X1(beta)
        self.f = f_models.SNorm(nvar)
        # na začatku nemáme vzorky - pouze rozdělení a podpís
        self.sample_box = SampleBox(self.f(), gm_signature=self.gm_signature)
        
        self.pf_exact = stats.norm.cdf(-beta)
        self.pf_exact_method = 'FORM (exact solution)' # Ang, Tang and Pythagoras
        self.r_exact = self.beta_exact
        
    def __str__(wt):
        return  'Line%sD' % wt.f.nvar
        
    def __repr__(wt):
        return  'Line(%s, %s)' % repr(wt.f.nvar, wt.beta_exact)
        
class TwoLine(WhiteBox):
    def __init__(self, nvar=2, beta=5):
        """
        Class takes for inicialization tuple of betas 
        Betas are coeffitients in sense of Regression Analysis (well, not really)
        g= a*X1 + b*X2 + c
        betas=(a,b,c)
        """
        self.beta = beta
        self.gm = g_models.AbsX1(beta)
        self.f = f_models.SNorm(nvar)
        # na začatku nemáme vzorky - pouze rozdělení a podpís
        self.sample_box = SampleBox(self.f(), gm_signature=self.gm_signature)
        
        self.pf_exact = stats.norm.cdf(-beta) * 2
        self.pf_exact_method = '2FORM (exact solution)' # Ang, Tang and Pythagoras
        self.r_exact = self.beta
        
        
    def __str__(wt):
        return  'TwoLine%sD' % wt.f.nvar
        
    def __repr__(wt):
        return  'TwoLine(%s, %s)' % repr(wt.f.nvar, wt.beta)


class Gaussian_Z_sum(WhiteBox): #ё куда ж без этого...
    def __init__(self, nvar=2, beta_exact=5):
        """
        
        """
        self.gm = g_models.Z_sum(nvar, beta_exact)
        self.f = f_models.SNorm(nvar)
        #č na začatku nemáme vzorky - pouze rozdělení a podpís
        self.sample_box = SampleBox(self.f(), gm_signature=self.gm_signature)
        
        self.beta_exact = beta_exact
        self.pf_exact = stats.norm.cdf(-self.beta_exact)
        self.pf_exact_method = 'FORM (exact solution)' # Ang, Tang and Pythagoras
        self.r_exact = self.beta_exact
        
        
    def __str__(wt):
        return  'Gaussian_Z_sum%sD'%(wt.f.nvar)
        
    def __repr__(wt):
        return  'Gaussian_Z_sum(nvar=%s, beta_exact=%s)' % (wt.f.nvar, wt.beta_exact)


def gauss_prod_CDF(z):
    """product Z= X_1 * X_2  has probability density f(z) = K_0(|z|)/pi, 
    where K_0() is the modified Bessel function of the second kind (zero order) and
    $X_1, X_2$ are independent standard normal variables.
    The distribution function F(z) = \int_{-\infty}^z f(t) dt  can be solved in closed form as follows
    F_Z(z)= 1/2 + z/2/pi * (K_0(|z|) * (2 + pi * L_1(z)) + K_1(|z|) * pi * L_0(|z|)),
    where L_0() L_1() are the modified Struve functions of orders zero and one respectively. 
    """
    StruveL0 = special.modstruve(0, np.abs(z))
    StruveL1 = special.modstruve(1, z)
    BesselK0 = special.kn(0, np.abs(z))
    BesselK1 = special.kn(1, np.abs(z))
    return 0.5 + z/2/np.pi * (BesselK0 * (2 + np.pi * StruveL1) + BesselK1 * np.pi * StruveL0)

class Gaussian_Z_prod_2D(WhiteBox): 
    def __init__(self, **kwargs):
        """
        Breitung RESS 182 (2019) p. 99
        """
        #č měníme logiku. 
        #č u této třídy známenko constanty bude ovlivňovat 
        #č poruchové kvadranty.
        if 'sign' in kwargs:
            self.sign = kwargs['sign']
        else:
            self.sign = 1
        if 'const' in kwargs:
            self.const = kwargs['const']
        elif 'beta' in kwargs:
            self.beta = kwargs['beta']
            self.const = self.beta**2/2
        else:
            raise ValueError
        
        
        #č g-modelu je to samozřejmě šuma, ale bílá skříňka nechť raději
        #č pečlivěji zpracovává vstup
        self.gm = g_models.Z_prod(const=self.const, sign=np.sign(self.sign))
        self.f = f_models.SNorm(2)
        #č na začatku nemáme vzorky - pouze rozdělení a podpís
        self.sample_box = SampleBox(self.f(), gm_signature=self.gm_signature)
        #č Rozdělení transformovaného náhodného čísla je zadano jako 
        # special.kn(0, np.abs(x)) / np.pi # Breitung RESS 182 (2019) p. 99
        # kn(n, x) Modified Bessel function of the second kind of integer order n
        # modstruve(v, x) Modified Struve function of order v at x
        #č Odvození pf_exact z Maple
        #self.pf_exact = 0.5 - const/2 * (StruveL1 * BesselK0 + StruveL0 * BesselK1 + 2/np.pi * BesselK0)
        if self.sign > 0:
            self.pf_exact = gauss_prod_CDF(-self.const)
        else:
            self.pf_exact = 1 - gauss_prod_CDF(-self.const)
        self.pf_exact_method = 'exact (Bessel) solution' 
        self.r_exact = np.sqrt(np.abs(self.const) * 2)
        
    def __str__(wt):
        return  'Gaussian_Z_prod_2D'
        
    def __repr__(wt):
        return  'Gaussian_Z_prod_2D(const=%s)' % wt.const



class Gaussian_Hyperbola_2D(WhiteBox): 
    def __init__(self, **kwargs):
        """
        Breitung RESS 182 (2019) p. 99
        """
        if 'const' in kwargs:
            self.const = kwargs['const']
        elif 'beta' in kwargs:
            self.beta = kwargs['beta']
            self.const = self.beta**2/2
        else:
            raise ValueError
        
        
        #č g-modelu je to samozřejmě šuma, ale bílá skříňka nechť raději
        #č pečlivěji zpracovává vstup
        self.gm = g_models.Z_hyperbola(self.const)
        self.f = f_models.SNorm(2)
        #č na začatku nemáme vzorky - pouze rozdělení a podpís
        self.sample_box = SampleBox(self.f(), gm_signature=self.gm_signature)
        #č Rozdělení transformovaného náhodného čísla je zadano jako 
        # special.kn(0, np.abs(x)) / np.pi # Breitung RESS 182 (2019) p. 99
        # kn(n, x) Modified Bessel function of the second kind of integer order n
        # modstruve(v, x) Modified Struve function of order v at x
        #č Odvození pf_exact z Maple
        #self.pf_exact = 0.5 - const/2 * (StruveL1 * BesselK0 + StruveL0 * BesselK1 + 2/np.pi * BesselK0)
        self.pf_exact = gauss_prod_CDF(-self.const) / 2
        self.pf_exact_method = 'exact (Bessel) solution' 
        self.r_exact = np.sqrt(np.abs(self.const) * 2)
        
    def __str__(wt):
        return  'Gaussian_Hyperbola_2D'
        
    def __repr__(wt):
        return  'Gaussian_Hyperbola_2D(const=%s)' % wt.const


#č já jsem si najednou uvědomil, že v tomto modulu
#č mám náprosto hrozný mix Camel- a snailcas'u.
#č Výmluvím z toho tak, podtržitko odděluje rozdělení a nazev g-modelu
class Gaussian_ProdFourBetas_2D(WhiteBox): 
    def __init__(self, beta=2):
        """
        Breitung RESS 182 (2019) p. 99
        """
        assert beta > 0
        self.beta = beta
        self.const = self.beta**2/2
        
        self.gm = g_models.Prod_FourBetas(self.beta)
        self.f = f_models.SNorm(2)
        #č na začatku nemáme vzorky - pouze rozdělení a podpís
        self.sample_box = SampleBox(self.f(), gm_signature=self.gm_signature)
        #č Rozdělení transformovaného náhodného čísla je zadano jako 
        # special.kn(0, np.abs(x)) / np.pi # Breitung RESS 182 (2019) p. 99
        # kn(n, x) Modified Bessel function of the second kind of integer order n
        # modstruve(v, x) Modified Struve function of order v at x
        #č Odvození pf_exact z Maple
        const = np.abs(self.const)
        StruveL0 = special.modstruve(0, const)
        StruveL1 = special.modstruve(1, const)
        BesselK0 = special.kn(0, const)
        BesselK1 = special.kn(1, const)
        self.pf_exact = 1 - const * (StruveL1 * BesselK0 + StruveL0 * BesselK1 + 2/np.pi * BesselK0)
        self.pf_exact_method = 'exact (Bessel) solution' 
        self.r_exact = self.beta
        
    def __str__(wt):
        return  'Gaussian_ProdFourBetas_2D'
        
    def __repr__(wt):
        return  'Gaussian_ProdFourBetas_2D(beta=%s)' % wt.beta


class Lognormal_Z_prod(WhiteBox): #č ověřím to moje odvození...
    def __init__(self, nvar=2, beta_exact=5, sign=1):
        """
        
        """
        self._sign = sign
        const = -np.exp(-np.sign(sign) * beta_exact * np.sqrt(nvar))
        self.gm = g_models.Z_prod(const, sign)
        self.f = f_models.UnCorD([stats.lognorm(s=1) for __ in range(nvar)])
        #č na začatku nemáme vzorky - pouze rozdělení a podpís
        self.sample_box = SampleBox(self.f(), gm_signature=self.gm_signature)
        
        self.beta_exact = beta_exact
        self.pf_exact = stats.norm.cdf(-self.beta_exact)
        self.pf_exact_method = 'FORM (exact solution)' # Ang, Tang and Pythagoras
        self.r_exact = self.beta_exact
        
    def __str__(wt):
        return  'Lognormal_Z_prod%sD'%(wt.f.nvar)
        
    def __repr__(wt):
        return  'Lognormal_Z_prod(nvar=%s, %s, %s)' % (wt.f.nvar, wt.beta_exact, wt.sign)


class Gaussian_Z_min(WhiteBox):
    def __init__(self, ndim=2, **kwargs):
        "je třeba zadat buď pf_exact, nebo konštantu u funkce minima Z_min"
        
        self.f = f_models.SNorm(ndim)
        self.ndim = ndim
    
        
        self.pf_exact_method = 'exact solution'
        if 'pf_exact' in kwargs:
            self.pf_exact = kwargs['pf_exact']
            sf = 1 - mpmath.root(1 - mpmath.mpf(self.pf_exact), ndim)
            self.const = stats.norm.isf(float(sf))
        elif 'const' in kwargs:
            self.const = kwargs['const']
            self.pf_exact = float(1 - mpmath.ncdf(self.const)**ndim)
        else:
            raise ValueError
        self.r_exact = self.const
        self.gm = g_models.Z_min(self.const) # min(X1, X2, XN) + const
        # na začatku nemáme vzorky - pouze rozdělení a podpís
        self.sample_box = SampleBox(self.f(), gm_signature=self.gm_signature)

    def __str__(self):
        return  'Gaussian_Z_min%sD'%(self.ndim)
        
    def __repr__(self):
        return  'Gaussian_Z_min(%s, pf_exact=%s)' % (repr(self.ndim), repr(self.pf_exact))


class Weibull_Z_min(WhiteBox):
    def __init__(self, wb_scales=(1.,1.), shape=5, **kwargs):
        """
        parametry pravdepodobnostniho rozdeleni pro Z_min s Weib. velicinami
        wb_scales=(1,1) - tuple of Weibull scale parameters, len(wb_scales)==nvar
        shape = 5 
        je třeba zadat buď pf_exact, nebo konštantu u funkce minima Z_min
        """
        self.wb_scales = wb_scales
        self.shape = 5
        self.f = f_models.UnCorD([stats.weibull_min(shape, scale=sc_i) for sc_i in wb_scales])
        
        # scale parametr minima z nvar Weibullovskych
        # tohle by platilo pro stejná rozdělení
        #sn = scale * nvar ** (-1.0 / shape)
        # pro nás musí to být něco takovýho
        sn = np.sum(np.power(wb_scales, -shape)) ** (-1.0 / shape)
        self.rvweibmin = stats.weibull_min(shape, scale=sn)
        
        # je třeba zadat buď pf_exact, nebo konštantu u funkce minima Z_min
        self.pf_exact_method = 'exact solution'
        if 'pf_exact' in kwargs:
            self.pf_exact = kwargs['pf_exact']
            self.const = -self.rvweibmin.ppf(self.pf_exact)
        elif 'beta_exact' in kwargs:
            self.beta_exact = kwargs['beta_exact']
            self.const = -self.rvweibmin.ppf(self.pf_exact)
        elif 'const' in kwargs:
            self.const = kwargs['const']
            self.pf_exact = self.rvweibmin.cdf(-self.const) # asi
        else:
            # no to teda uživatele пошли
            self.pf_exact = 1e-4
            self.const = -self.rvweibmin.ppf(self.pf_exact)
        
        sf = 1 - mpmath.root(1 - mpmath.mpf(self.pf_exact), len(wb_scales))
        self.r_exact = stats.norm.isf(float(sf))
        self.gm = g_models.Z_min(self.const)
        # na začatku nemáme vzorky - pouze rozdělení a podpís
        self.sample_box = SampleBox(self.f(), gm_signature=self.gm_signature)

    def __str__(self):
        return  'Weibull_Z_min%sD'%(len(self.wb_scales))
        
    def __repr__(self):
        return  'Weibull_Z_min(%s, %s, pf_exact=%s)' % (repr(self.wb_scales), repr(self.shape), repr(self.pf_exact))


class Gaussian_Z_max(WhiteBox):
    def __init__(self, ndim=2, **kwargs):
        "je třeba zadat buď pf_exact, nebo konštantu u funkce maxima Z_max"
        
        self.f = f_models.SNorm(ndim)
        self.ndim = ndim
    
        
        self.pf_exact_method = 'exact solution'
        if 'pf_exact' in kwargs:
            self.pf_exact = kwargs['pf_exact']
            cdf = mpmath.root(self.pf_exact, ndim)
            self.const = -stats.norm.ppf(float(cdf))
        elif 'const' in kwargs:
            self.const = kwargs['const']
            self.pf_exact = float(mpmath.ncdf(-self.const)**ndim)
        else:
            raise ValueError
        self.r_exact = np.sqrt(self.const**2 * ndim)
        self.gm = g_models.Z_max(self.const) # min(X1, X2, XN) + const
        # na začatku nemáme vzorky - pouze rozdělení a podpís
        self.sample_box = SampleBox(self.f(), gm_signature=self.gm_signature)

    def __str__(self):
        return  'Gaussian_Z_max%sD'%(self.ndim)
        
    def __repr__(self):
        return  'Gaussian_Z_max(%s, pf_exact=%s)' % (repr(self.ndim), repr(self.pf_exact))


#č masturbace
class Gaussian_Z_sumexp_2D(WhiteBox):
    def __init__(self, const):
        self.const = const
        self.f = f_models.SNorm(2)
        self.beta = np.sqrt(-np.log(-self.const/2)*2) 
        self.pf_exact_method = 'rude (product) asymptotic approximation'
        self.pf_exact = stats.norm.cdf(-self.beta) * 2 * np.sqrt(2)
        
        self.gm = g_models.Z_sumexp(self.const)
        # na začatku nemáme vzorky - pouze rozdělení a podpís
        self.sample_box = SampleBox(self.f(), gm_signature=self.gm_signature)
     
    def __str__(self):
        return  'Gaussian_Z_sumexp_2D'
        
    def __repr__(self):
        return  'Gaussian_Z_sumexp_2D(%s)' % repr(self.const)

#č masturbace pro příště
#č tady jsou nekorektní hračky s momenty rozdělení
#č není známo, jestli se někdy do tohoto kódu dorazí
#č korektní výpočet pf
#class Gaussian_Z_sumexp(WhiteBox):
#    def __init__(self, nvar=2, **kwargs):
#        """
#        je třeba zadat buď pf_exact, nebo konštantu u funkce Z_sumexp
#        """
#        
#        # je tam předpoklad SNormu?
#        self.f = f_models.SNorm(nvar)
#        
#        self.C1 = np.sqrt(np.sqrt(5) / 3. - 1. / 3.)
#        self.C2 = np.sqrt(3.) / 3.
#        # je třeba zadat buď pf_exact, nebo konštantu u funkce Z_sumexp
#        self.pf_exact_method = 'tešt solution'
#        if 'pf_exact' in kwargs:
#            self.pf_exact = kwargs['pf_exact']
#            self.C = self.beta_exact * self.C1 * np.sqrt(nvar) - self.C2 * nvar
#        elif 'beta_exact' in kwargs:
#            self.beta_exact = kwargs['beta_exact']
#            self.C = self.beta_exact * self.C1 * np.sqrt(nvar) - self.C2 * nvar
#        elif 'const' in kwargs:
#            self.const = kwargs['const']
#            self.C = self.const
#            self.beta_exact = (self.C + self.C2 * nvar) / self.C1 / np.sqrt(nvar)
#        elif 'C' in kwargs:
#            self.C = kwargs['C']
#            self.beta_exact = (self.C + self.C2 * nvar) / self.C1 / np.sqrt(nvar)
#        else:
#            # no to teda uživatele пошли
#            self.pf_exact = 1e-4
#            self.C = self.beta_exact * self.C1 * np.sqrt(nvar) - self.C2 * nvar
#        
#        
#        self.const = self.C
#        self.gm = g_models.Z_sumexp(self.const)
#        # na začatku nemáme vzorky - pouze rozdělení a podpís
#        self.sample_box = SampleBox(self.f(), gm_signature=self.gm_signature)
#     
#    def __str__(self):
#        return  'Gaussian_Z_sumexp%sD'%(self.nvar)
#        
#    def __repr__(self):
#        return  'Gaussian_Z_sumexp(%s, pf_exact=%s)' % (repr(self.nvar), repr(self.pf_exact))   



class SNorm_Z_sumsq(WhiteBox):
    def __init__(self, nvar=2, **kwargs):
        """
        je třeba zadat buď pf_exact, nebo konštantu u funkce Z_sumsq
        """
        
        # je tu předpoklad SNormu, to vím
        self.f = f_models.SNorm(nvar)
        
        self.rvchisq = stats.chi2(nvar)
        
        # je třeba zadat buď pf_exact, nebo konštantu u funkce Z_sumsq
        self.pf_exact_method = 'exact solution'
        if 'pf_exact' in kwargs:
            self.pf_exact = kwargs['pf_exact']
            self.C = self.rvchisq.ppf(self.pf_exact)
        elif 'beta_exact' in kwargs:
            self.beta_exact = kwargs['beta_exact']
            self.C = self.rvchisq.ppf(self.pf_exact)
        elif 'const' in kwargs:
            self.const = kwargs['const']
            self.C = self.const
            self.pf_exact = self.rvchisq.cdf(self.C)
        elif 'C' in kwargs:
            self.C = kwargs['C']
            self.pf_exact = self.rvchisq.cdf(self.C)
        else:
            # no to teda uživatele пошли
            self.pf_exact = 1e-4
            self.C = self.rvchisq.ppf(self.pf_exact)
        
        
        self.const = self.C
        self.R_exact = np.sqrt(self.const)
        self.gm = g_models.Z_sumsq(self.C)
        # na začatku nemáme vzorky - pouze rozdělení a podpís
        self.sample_box = SampleBox(self.f(), gm_signature=self.gm_signature)
     
    def __str__(self):
        return  'SNorm_Z_sumsq%sD'%(self.nvar)
        
    def __repr__(self):
        return  'SNorm_Z_sumsq(%s, pf_exact=%s)' % (repr(self.nvar), repr(self.pf_exact))   






class SNorm_S_ball(WhiteBox):
    #r = 5.256521 # pf 1.00000404635e-06
    def __init__(self, nvar=2, r=5.256521):
        
        
        # SNorm
        self.f = f_models.SNorm(nvar)
        
        
        #
        #   pf, jen tak, hračka
        #
        self.pf_exact_method = 'precise solution'
        if nvar == 1:
            #self.pf_exact = 1 - 2**(1-nvar/2) / special.gamma(nvar/2)    *    (np.sqrt(np.pi)*special.erf(r/np.sqrt(2)))/np.sqrt(2)
            self.pf_exact = 1 - special.erf(r/1.4142135623730951)
        elif nvar == 2:
            self.pf_exact = np.exp(-r**2/2)
        elif nvar == 3:
            #self.pf_exact = 1 - 2**(1-nvar/2) / special.gamma(nvar/2)    *    (np.exp(-r**2/2)*(np.sqrt(np.pi)*np.exp(r**2/2)*special.erf(r/np.sqrt(2))-np.sqrt(2)*r))/np.sqrt(2)
            self.pf_exact = 1 - 0.5641895835477564 * (np.exp(-r**2/2)*(np.sqrt(np.pi)*np.exp(r**2/2)*special.erf(r/np.sqrt(2))-np.sqrt(2)*r))
        elif nvar == 4:
            self.pf_exact = (r**2/2+1)*np.exp(-r**2/2)
        elif nvar == 6:
            self.pf_exact = (r**4+4*r**2+8)*np.exp(-r**2/2)/8
            
            # nvar=8:  (48-(r^6+6*r^4+24*r^2+48)*e^(-r^2/2)  / 2**(nvar/2))/48
            
            # hračička ve hračce
            # nemám žádnou jistotu, že tohle počítá přesněji
        elif nvar % 2 == 0: # sudé
            poly = [1]
            for i in range(nvar-2, 0, -2):
                poly.append(0)
                poly.append(i*poly[-2])
            self.pf_exact = np.polyval(np.array(poly) / poly[-1], r) * np.exp(-r**2/2) 
            
        else:
            self.pf_exact = 1 - 2**(1-nvar/2) / special.gamma(nvar/2)    *    integrate.quad(lambda x: np.exp(-(x**2)/2)*x**(nvar-1), 0, r)[0] 
        
        
        self.r = r
        self.r_exact = self.r
        self.gm = g_models.S_ball(r)
        # na začatku nemáme vzorky - pouze rozdělení a podpís
        self.sample_box = SampleBox(self.f(), gm_signature=self.gm_signature)
     
    def __str__(self):
        return  'SNorm_S_ball%sD'%(self.nvar)
        
    def __repr__(self):
        return  'SNorm_S_ball(nvar=%s, r=%s)' % (repr(self.nvar), repr(self.r))   





#!/usr/bin/env python
# coding: utf-8


#č nazvy proměnných jsou v angličtině
#č Ale komenty teda ne)

import numpy as np
import matplotlib.tri as mtri
import matplotlib.path as mpath
import matplotlib.colors as mcolor
import matplotlib.patches as mpatches

from .. import misc as wmisc

#č Tahlensta blbost je použita funkcí tripcolor()
#č Je třeba jú překopat na Triangulation třidu,
#č která get_events má jako svůj method.
def get_events(sb, simplices): #simplices = bx.tri.simplices
    """
    Metoda musí simplexům přiřazovat jev 
    0=success, 1=failure, 2=mix
    """
    in_failure = np.isin(simplices, sb.failure_points)
    has_failure = in_failure.any(axis=1)
    all_failure = in_failure.all(axis=1)
    return np.int8(np.where(has_failure, np.where(all_failure, 1, 2), 0))

#č napadlo mě, že bych mohl matplotlibovskému Axes
#č přiřazovat (připsavat, zadávat) atribut 'space'
#č Daválo by to smysl, ne? U všeho ostatního, u sample boksů
#č ne vždy na jedném sabplotu někdo potřebuje
#č kreslit z různejch prostoru.
#č Zkrátka, funkce v tomto modulu požadujou aby 
#č ax.space a ax.sample_box byl nastaven!

# ax.space and ax.sample_box attributes should (expected to) be set up!


def scatter_sample(ax, sample, **kwargs):
    xy = getattr(sample, ax.space)
    x, y = xy[:,:2].T
    return ax.scatter(x, y, **kwargs)

def plot_sample(ax, sample, *args, **kwargs):
    xy = getattr(sample, ax.space)
    x, y = xy[:,:2].T
    return ax.plot(x, y, *args, **kwargs)



def scatter_points(ax, **kwargs):
    xy = getattr(ax.sample_box, ax.space)
    nsim = len(xy)
    x, y = xy[:,:2].T
    
    failsi = ax.sample_box.failsi
    
    try: # proxy denotes to implicitly-known values
        proxy = ax.sample_box.proxy.astype(bool) 
    except AttributeError:
        proxy = np.full(nsim, False, dtype=bool)
    
    scatter_list = []
    
    #č byl jsem svědkem, že matplotlib zlobil ve 3D 
    #č kvůli tomu, že nebyl žádný safe vzorek
    #č proto raději budu přidávat tečky podmíněne
    mask = np.all((~failsi, ~proxy), axis=0)
    if np.any(mask): #success
        scatter_list.append(ax.scatter(x[mask], y[mask], c='g', marker='P', **kwargs))
    
    mask = np.all((failsi, ~proxy), axis=0)
    if np.any(mask): #failures
        scatter_list.append(ax.scatter(x[mask], y[mask], c='r', marker='X', **kwargs))
    
    mask = np.all((~failsi, proxy), axis=0)
    if np.any(mask): #proxy_successes
        scatter_list.append(ax.scatter(x[mask], y[mask], c='#77AC30', marker='h', **kwargs))
    
    mask = np.all((failsi, proxy), axis=0)
    if np.any(mask): #proxy_failures
        scatter_list.append(ax.scatter(x[mask], y[mask], c='#D95319', marker='H', **kwargs))
    
    return scatter_list # success, failures, proxy_successes, proxy_failures
    

def plot_points(ax, ls='', **kwargs):
    xy = getattr(ax.sample_box, ax.space)
    nsim = len(xy)
    x, y = xy[:,:2].T
    
    failsi = ax.sample_box.failsi
    
    try: # proxy denotes to implicitly-known values
        proxy = ax.sample_box.proxy.astype(bool) 
    except AttributeError:
        proxy = np.full(nsim, False, dtype=bool)
    
    plot_list = []
    
    #č byl jsem svědkem, že matplotlib zlobil ve 3D 
    #č kvůli tomu, že nebyl žádný safe vzorek
    #č proto raději budu přidávat tečky podmíněne
    mask = np.all((~failsi, ~proxy), axis=0)
    if np.any(mask): #success
        plot_list.append(ax.plot(x[mask], y[mask], mec='g', mfc='g',\
                         marker='P', ls=ls, **kwargs))
    
    mask = np.all((failsi, ~proxy), axis=0)
    if np.any(mask): #failures
        plot_list.append(ax.plot(x[mask], y[mask], mec='r', mfc='r',\
                         marker='X', ls=ls, **kwargs))
    
    mask = np.all((~failsi, proxy), axis=0)
    if np.any(mask): #proxy_successes
        plot_list.append(ax.plot(x[mask], y[mask], mec='#77AC30', mfc=(0,0,0,0),\
                         marker='h', ls=ls, **kwargs))
    
    mask = np.all((failsi, proxy), axis=0)
    if np.any(mask): #proxy_failures
        plot_list.append(ax.plot(x[mask], y[mask], mec='#D95319', mfc=(0,0,0,0),\
                         marker='H', ls=ls, **kwargs))
    
    return plot_list # success, failures, proxy_successes, proxy_failures


#č triplot - pokud ax.space == tri_space
#č tri_plot - pokud ax.space != Tri.tri_space
def tri_plot(ax, Tri=None, fmt='-', ns=100, **kwargs):
    if ax.sample_box.nvar == 2:
        lines = []
        
        if Tri is None:
            Tri = ax.sample_box.Tri
        
        # take coordinates in the space, where triangulation has been performed
        sampled_plan_tri = getattr(ax.sample_box, Tri.tri_space)
        
        #č mohli bychom zde machrovat se zkracenou smyčkou,
        #č mohli bychom jednoduše zavolat zabudovanou v matplotlib funkciju,
        #č my ale vůbec kontrolovat rovnost prostorů nebudeme, 
        #č nechť to dělá volající kód
        
        #if ax.space == Tri.tri_space:
        #   #return triplot(ax, **kwargs)
        #   
        #    #for simplex_id in simplex_ids:
        #    #    triangle = simplices[simplex_id]
        #    #    pos = sampled_plan_tri[triangle[[0,1,2,0]]]
             
        for triangle in Tri.tri.simplices:
            x_tri_1 = np.linspace(sampled_plan_tri[triangle[0],0], sampled_plan_tri[triangle[1],0], ns, endpoint=False)
            y_tri_1 = np.linspace(sampled_plan_tri[triangle[0],1], sampled_plan_tri[triangle[1],1], ns, endpoint=False)
            x_tri_2 = np.linspace(sampled_plan_tri[triangle[1],0], sampled_plan_tri[triangle[2],0], ns, endpoint=False)
            y_tri_2 = np.linspace(sampled_plan_tri[triangle[1],1], sampled_plan_tri[triangle[2],1], ns, endpoint=False)
            x_tri_3 = np.linspace(sampled_plan_tri[triangle[2],0], sampled_plan_tri[triangle[0],0], ns, endpoint=True)
            y_tri_3 = np.linspace(sampled_plan_tri[triangle[2],1], sampled_plan_tri[triangle[0],1], ns, endpoint=True)
            
            tri_bound_tri = np.concatenate(((x_tri_1, y_tri_1), (x_tri_2, y_tri_2), (x_tri_3, y_tri_3)), axis=1).T
            #č vytvořme sample
            tri_bound = ax.sample_box.f_model.new_sample(tri_bound_tri, space=Tri.tri_space)
            
            xy = getattr(tri_bound, ax.space)
            x, y = xy.T
            lines.append(ax.plot(x, y, fmt, **kwargs))
        
        return lines


#č triplot() jednoduše volá zabudovanou do matplotlibu funkci
def triplot(ax, **kwargs):
    xy = getattr(ax.sample_box, ax.space)
    x, y = xy[:,:2].T
    
    return ax.triplot(x, y, **kwargs)


def tripcolor(ax, sfm_colors=None, **kwargs):
    xy = getattr(ax.sample_box, ax.space)
    x, y = xy[:,:2].T
    tri = mtri.Triangulation(x, y)
    
    if sfm_colors is None:
        # make a color map of fixed colors
        s = '#a7ffb5' #'xkcd:light seafoam green' #a7ffb5
        f = '#fdc1c5' #'xkcd: pale rose' # (#fdc1c5)
        m = '#FFF39A' #'xkcd: dark cream' # (255, 243, 154, 255)
        sfm_colors = [s, f, m]
    
    if 'cmap' not in kwargs:
        # 0=success, 1=failure, 2=mix
        kwargs['cmap'] = mcolor.ListedColormap(sfm_colors)
    if 'norm' not in kwargs:
        kwargs['norm'] = mcolor.NoNorm()
    
    # same as facecolors
    C = get_events(ax.sample_box, tri.get_masked_triangles())
    
    #č tak to má bejt, aby MPL jednoznačně bral barvy jako barvy obličejů
    #č jenomže to může zlobit
    return ax.tripcolor(tri, facecolors=C, **kwargs)
    

def plot_boundaries(ax, fmt='-b', nrod=200, **kwargs):
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    limits = np.array([[xmin, ymin], [xmax, ymax]])
    bounds = ax.sample_box.get_2D_boundary(nrod=nrod, viewport_sample=limits,\
                                             viewport_space=ax.space)
    lines = []
    for bound in bounds:
        xy = getattr(bound, ax.space)
        x, y = xy[:,:2].T
        lines.append(ax.plot(x, y, fmt, **kwargs))
    
    return lines



def scatter_candidates(ax, **kwargs):
    """
    Plot all nodes series from ax.sample_box.candidates_index list.
    Function extracts ax.sample_box.potential attribute from nodes 
    and uses it for colormapping. 
    Max value is taken from ax.sample_box.highest_bid
    Returns nothing.
    example:
    scatter_candidates(ax, s=100.500, marker='.', cmap='plasma', alpha=0.5, linewidths=None, *, edgecolors=None, plotnonfinite=False)
    """
    potential = ax.sample_box.potential
    maxcb = ax.sample_box.highest_bid
    
    #č a teď jdeme!
    for id, cb in ax.sample_box.candidates_index.items():
        values = getattr(cb, potential)
        x, y = getattr(cb, ax.space)[:,:2].T
        ax.scatter(x, y, c=values, vmin=0, vmax=maxcb, **kwargs)


def plot_the_best_candidate(ax, *args, **kwargs):
    """
    Plots ax.sample_box.bidder node.
    Returns nothing.
    example:
    plot_the_best_candidate(ax, "^", color='green')
    """
    xy = getattr(ax.sample_box.bidder, ax.space)
    x, y = xy[:,:2].T
    return ax.plot(x, y, *args, **kwargs)



def simplex_vectors(ax, **kwargs):
    Tri = ax.sample_box.Tri
    if ax.space != Tri.tri_space:
        raise
    
    result = Tri.perform_sensitivity_analysis()
    vectors = result.vectors
    probabilities = result.shares
    
    ED = Tri.tri.points
    simplices = Tri.tri.simplices
    #pos = getattr(self.w.sample_box, self.w.space)[:,:2]
    
#    offsets = {}
#    for key, vector in result.unique_id_vectors.items():
#        offsets[key] = np.mean(Tri.get_finalized_supports(vector))
        
    
    X = []
    Y = []
    U = []
    V = []
    for simplex_id, vector in vectors.items():
        centroid = np.mean(ED[simplices[simplex_id]], axis=0)
        #offset_fix = (offsets[id(vector)] - np.inner(centroid, vector.normal))
        #placement = centroid + vector.normal * offset_fix
        x, y = centroid #placement
        X.append(x)
        Y.append(y)
        u, v = -vector.normal * probabilities[simplex_id]
        U.append(u)
        V.append(v)
    
    
    return ax.quiver(X, Y, U, V, **kwargs)



def qhull_polygon(ax, qhull, **kwargs):
    x, y = qhull.points[qhull.vertices].T
    return ax.fill(x, y, **kwargs)


                
def qhull_plot(ax, qhull=None, fmt='-', ns=100, **kwargs):
    if ax.sample_box.nvar == 2: #č jinak nic nedeláme
        if qhull is None:
            qhull = ax.sample_box.convex_hull
            
        if ax.space == qhull.space:
            points = qhull.points
            lines = []
            for simplex in qhull.simplices:
                xy = points[simplex]
                x, y = xy.T
                lines.append(ax.plot(x, y, fmt, **kwargs))
            return lines
        
        else:
            #оӵ кулэ ӧвӧл обновлять экран карыны
            sampled_plan_tri = qhull.points
            for simplex in qhull.simplices:
                start_id, end_id = simplex
                
                x_bound = np.linspace(sampled_plan_tri[start_id,0], sampled_plan_tri[end_id,0], ns, endpoint=True)
                y_bound = np.linspace(sampled_plan_tri[start_id,1], sampled_plan_tri[end_id,1], ns, endpoint=True)
                
                # sample compatible
                #оӵ малы транспонировать кароно? Озьы кулэ!
                bound_tri = np.vstack((x_bound, y_bound)).T
                #č vytvořme sample
                bound = ax.sample_box.f_model.new_sample(bound_tri, space=qhull.space)
                
                xy = getattr(bound, ax.space)
                x, y = xy.T
                lines.append(ax.plot(x, y, fmt, **kwargs))
                
            return lines



def dhull_plot(ax, hull, **kwargs):
    #č zatím uděláme jen pro 2D infinite lajny
    design_points = hull.get_design_points()
    lines = []
    if (hull.sample.nvar == 2) and (ax.space == hull.space):
        for equation in hull.equations:
            #č ve 2D bych očekával v rovnici pouze 3 hodnoty (já potřebuji směry)
            x, y, offset = equation
            design_point = [-x*offset, -y*offset] #č offset je prej zápornej
            slope = np.divide(-x, y)
            lines.append(ax.axline(design_point, slope=slope, **kwargs))
    return lines
            
def bhull_plot(ax, bhull, **kwargs):
    if ax.space == bhull.space:
        point = bhull.mins[:2]
        x1, y1 = point
        x2, y2 = bhull.maxs[:2]
        if 'fill' not in kwargs:
            kwargs['fill'] = False
        frame = mpatches.Rectangle(point, x2-x1, y2-y1, **kwargs)
        return ax.add_patch(frame) 

def shull_plot(ax, hull, **kwargs):
    from ..ghull import Ghull
    ghull = Ghull(hull)
    R = ghull.get_R()
    if 'fill' not in kwargs:
        kwargs['fill'] = False
    return gcircle(ax, r=R, **kwargs)


## DEPRECATED
## use qhull_plot instead
#def convex_plot(ax, fmt='-m', ns=100, qhull=None, **kwargs):
#    if ax.sample_box.nvar == 2: #č jinak nic nedeláme
#       if qhull is None:
#           simplices = ax.sample_box.convex_hull.simplices
#        else:
#           simplices = qhull.simplices
#        
#        if tri_space is None:
#           
#        # convex hull should be made in the same space as triangulation, 
#        # Will we take coordinates in the triangulation space, I guess?
#        sampled_plan_tri = getattr(ax.sample_box, ax.sample_box.tri_space)
#        
#        # hmm...
#        lines = []
#        if ax.space == ax.sample_box.tri_space:
#            for simplex in simplices:
#                xy = sampled_plan_tri[simplex]
#                x, y = xy.T
#                lines.append(ax.plot(x, y, fmt, **kwargs))
#                    
#    else:
#        #оӵ кулэ ӧвӧл обновлять экран карыны
#        for simplex in simplices:
#            start_id, end_id = simplex
#            
#            x_bound = np.linspace(sampled_plan_tri[start_id,0], sampled_plan_tri[end_id,0], ns, endpoint=True)
#            y_bound = np.linspace(sampled_plan_tri[start_id,1], sampled_plan_tri[end_id,1], ns, endpoint=True)
#            
#            # sample compatible
#            #оӵ малы транспонировать кароно? Озьы кулэ!
#            bound_tri = np.vstack((x_bound, y_bound)).T
#            #č vytvořme sample
#            bound = ax.sample_box.f_model.new_sample(bound_tri, space=ax.sample_box.tri_space)
#            
#            xy = getattr(bound, ax.space)
#            x, y = xy.T
#            lines.append(ax.plot(x, y, fmt, **kwargs))
#                
#                
#    return lines



def rbf_density_colormesh(ax, ngrid=500, **kwargs):
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    
    x = np.linspace(xmin, xmax, ngrid, endpoint=True)
    y = np.linspace(ymin, ymax, ngrid, endpoint=True)
    X, Y = np.meshgrid(x, y)
    
    XY = np.vstack((X.reshape(-1), Y.reshape(-1))).T
    z = ax.sample_box.sample_pdf(XY, ax.space)
    Z = z.reshape(ngrid, ngrid)
    
    #č s tou alphou mně to nějak nepovedlo
    #č matplotlib ji nějak ignoruje 
    #č a podle toho, co vidím v kódu
    #č není to problém kolormapy LinearSegmentedColormap
#    cdict = {'red':[[0.0,  253/255, 253/255],
#                   [0.9999,  253/255, 167/255],
#                   [1.0,  167/255, 167/255]],
#         'green': [[0.0,  193/255, 193/255],
#                   [0.9999,  193/255, 1.0],
#                   [1.0,  1.0, 1.0]],
#         'blue':  [[0.0,  197/255, 197/255],
#                   [0.9999,  197/255, 181/255],
#                   [1.0,  181/255, 181/255]],
#         'alpha':  [[0.0,  1.0, 1.0],
#                   [0.9999,  1.0, 0.2],
#                   [1.0,  0.2, 0.2]]}
    
    cdict = {'red':[[0.0,  1.0, 1.0],
                   [0.99999,  253/255, 167/255],
                   [1.0,  167/255, 167/255]],
         'green': [[0.0,  1.0, 1.0],
                   [0.99999,  0/255, 1],
                   [1.0,  1.0, 1.0]],
         'blue':  [[0.0,  1.0, 1.0],
                   [0.99999,  17/255, 181/255],
                   [1.0,  181/255, 181/255]]}
                   

    cmap = mcolor.LinearSegmentedColormap('red_density', segmentdata=cdict, N=2560)
    #cmap = mcolors.ListedColormap(['#A7FFB5', '#FDC1C5'])
    rbf_values = wmisc.RBF_surrogate(ax.sample_box, ax.space).rbf(X, Y)
    #Z = np.log(Z/np.max(Z[rbf_values <= 0]))
    #C = -Z/np.min(Z) * (rbf_values <= 0) + 1/250 * (rbf_values > 0)
    Z = Z / np.max(Z[rbf_values <= 0])
    C = (np.sqrt(Z)-1) * (rbf_values <= 0) + 1/250 * (rbf_values > 0)
    return ax.pcolormesh(X, Y, C, cmap=cmap, shading='nearest', edgecolors='face',
                        zorder=-100, rasterized=True, **kwargs)
    

def rbf_colormesh(ax, ngrid=500, **kwargs):
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    
    x = np.linspace(xmin, xmax, ngrid, endpoint=True)
    y = np.linspace(ymin, ymax, ngrid, endpoint=True)
    X, Y = np.meshgrid(x, y)

    cmap = mcolor.ListedColormap(['#A7FFB5', '#FDC1C5'])
    rbf_values = wmisc.RBF_surrogate(ax.sample_box, ax.space).rbf(X, Y)
    C = rbf_values < 0
    return ax.pcolormesh(X, Y, C, cmap=cmap, shading='nearest', zorder=-100, rasterized=True, **kwargs)



def gcircle(ax, r=1, nrod=200, **kwargs):
    if ax.space == 'G':
        circle = mpatches.Circle((0,0), r, **kwargs)
    else:
        sample = wmisc.get_isodistances(ax.sample_box.f_model, r, nrod=200)
        xy = getattr(sample, ax.space)[:,:2]
        circle = mpatches.Polygon(xy, **kwargs)
    
    #č vrací add_patch něco?
    return ax.add_patch(circle)


def uframe(ax, **kwargs):
    if ax.space in ('P', 'aP', 'U', 'aU'):
        alpha = ax.sample_box.alpha
        frame = mpatches.Rectangle((0,0), alpha[0], alpha[1], fill=False, **kwargs)
        return ax.add_patch(frame) 


def lsf_boundary(ax, ngrid=200, limits=None, **kwargs):
        
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
        
    if limits is not None: # G-čko zlobí
        lxmin, lxmax, lymin, lymax = limits
        xmin = min(xmin, lxmin)
        xmax = max(xmax, lxmax)
        ymin = min(ymin, lymin)
        ymax = max(ymax, lymax)
    
    x = np.linspace(xmin, xmax, ngrid)
    y = np.linspace(ymin, ymax, ngrid)
    X, Y = np.meshgrid(x, y)
    XY = np.vstack((X.reshape(-1), Y.reshape(-1))).T
    sb = ax.sample_box.gm(ax.sample_box.f_model.new_sample(XY, space=ax.space))
    Z = sb.g_values
    return ax.contour(X, Y, Z.reshape(ngrid, ngrid), np.zeros(1), **kwargs)


def isocurves(ax, ngrid=200, limits=None, ncurves=5, **kwargs):
    r = ncurves + 1 #č jakási rezerva, mám už plné zuby s těmi limity
    sample = wmisc.get_isodistances(ax.sample_box.f_model, r, nrod=ngrid)
    xy = getattr(sample, ax.space)[:,:2]
        
    xmin, ymin = np.min(xy, axis=0)
    xmax, ymax = np.max(xy, axis=0)
        
    if limits is not None: # G-čko zlobí
        lxmin, lxmax, lymin, lymax = limits
        xmin = min(xmin, lxmin)
        xmax = max(xmax, lxmax)
        ymin = min(ymin, lymin)
        ymax = max(ymax, lymax)
    
    x = np.linspace(xmin, xmax, ngrid)
    y = np.linspace(ymin, ymax, ngrid)
    X, Y = np.meshgrid(x, y)
    XY = np.vstack((X.reshape(-1), Y.reshape(-1))).T
    Z = ax.sample_box.f_model.sample_pdf(XY, ax.space)
    
    
    const = 1 / (xmax - xmin) / (ymax - ymin)
    r_levels = np.arange(ncurves) + 1
    levels = wmisc.isolevels_2d(Z, const, np.flip(r_levels), from_top=False)
    return ax.contour(X, Y, Z.reshape(ngrid, ngrid), levels, **kwargs)



def number_points(ax, **kwargs):
    points_coordinates = getattr(ax.sample_box, ax.space)[:,:2]
    x, y = points_coordinates.T
    labels = []
    for i in range(ax.sample_box.nsim):
        labels.append(ax.text(x[i], y[i], str(i+1), **kwargs))
    return labels


def center_spines(ax, lw=0.4):
    # Move the left and bottom spines to x = 0 and y = 0, respectively.
    ax.spines["left"].set_position(("data", 0))
    ax.spines["bottom"].set_position(("data", 0))
    
    # Show the left and bottom spines
    ax.spines["left"].set_visible(True)
    ax.spines["bottom"].set_visible(True)
    
    # Set up linewidth
    ax.spines["left"].set_linewidth(lw)
    ax.spines["bottom"].set_linewidth(lw)
    
    # Hide the top and right spines.
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    
    # instead of black triangles ">k"/"^k"
    vertices = np.array([[0, 1], [5, 0], [0, -1], [0, 1]])
    right_arrow = {
            'marker': mpath.Path(vertices, codes=None, closed=True),
            'markerfacecolor': 'black',
            'markeredgecolor': 'black',
            'markersize': 15
            }
    
    vertices = np.array([[-1, 0], [0, 5], [1, 0], [-1, 0]])
    up_arrow = {
            'marker': mpath.Path(vertices, codes=None, closed=True),
            'markerfacecolor': 'black',
            'markeredgecolor': 'black',
            'markersize': 15
            }
    
    # Draw arrows (as markers) at the end of the axes.  In each
    # case, one of the coordinates (0) is a data coordinate (i.e., y = 0 or x = 0,
    # respectively) and the other one (1) is an axes coordinate (i.e., at the very
    # right/top of the axes).  Also, disable clipping (clip_on=False) as the marker
    # actually spills out of the axes.
    ax.plot(1, 0, **right_arrow, transform=ax.get_yaxis_transform(), clip_on=False)
    ax.plot(0, 1, **up_arrow, transform=ax.get_xaxis_transform(), clip_on=False)

#č před použitím bude třeba skutečně naimportovat ticker
def fix_locator(ax, loc):
    loc_x = mpl.ticker.FixedLocator(loc)
    ax.xaxis.set_major_locator(loc_x)
    loc_y = mpl.ticker.FixedLocator(loc)
    ax.yaxis.set_major_locator(loc_y)

#č něco_kulatého
def curly(ax, linewidths=[0.7, 0.5, 0.4, 0.3, 0.2, 0.1], nrid=200, color='k', **kwargs):
    if ax.space in ('U', 'aU'):
        return None
        
    elif (ax.sample_box.nvar == 2) and (ax.space in ('Rn', 'aRn', 'R', 'aR', 'P', 'aP')):
        limits = (*ax.get_xlim(), *ax.get_ylim())
        isocurves(ax, ngrid=nrid, limits=limits, ncurves=len(linewidths),\
                     linewidths=np.flip(linewidths), colors=[color], **kwargs)
        
    else:
        for i, lw in zip(range(len(linewidths)), linewidths):
            gcircle(ax, r=i+1, nrod=nrid, color=color, linewidth=lw, fill=False)


def setup(ax, lw=0.4): #č šup
    #ax.set_xlabel('$x_{1}$')
    #ax.set_ylabel('$x_{2}$')
    
    #ax.set_frame_on(False) #č pak se mi nezobrazí osy
    ax.set_aspect(1)
    #ax.set_box_aspect(1)
    if ax.space in ('P', 'aP', 'U', 'aU'):
        ax.margins(0)
        #ax.set_frame_on(False)
        #uframe(ax, linewidth=lw) 
    else:
        center_spines(ax, lw)


def setup_labels(ax, x_label="$x_1$", y_label="$x_2~$", data_offset=1):
    if ax.space in ('P', 'aP', 'U', 'aU'):
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)
    else:
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        text = ax.text(1, -data_offset / np.abs(ymin), x_label, ha='center',va='top',
                                             transform=ax.get_yaxis_transform())
        text.set_in_layout(False)
        text = ax.text(-data_offset / np.abs(xmin), 1, y_label, ha='right',va='center', 
                                        transform=ax.get_xaxis_transform())
        text.set_in_layout(False)





#!/usr/bin/env python
# coding: utf-8

import numpy as np


#č v statistice rozptyl jednotlyvých měření charakterizuje přesnost
# https://en.wikipedia.org/wiki/Inverse-variance_weighting
#č z teorii chyb měření
#č mame-li změřené dvě části l_1 a l_2, co musejí dohromady dat L,
#č můžeme spočítat odchylku delta = L - l_1 - l_2
#č Znamé-li přesnost jednotlivých měření 1/var_1 a 1/var_2,
#č můžeme tvrdit, že opravy jsou:
# v_1 = delta * var_1 / (var_1 + var_2)
# v_2 = delta * var_2 / (var_1 + var_2)
#č a tedy, že výrovnáné hodnoty jsou:
# l_1' = l_1 + v_1
# l_2' = l_2 + v_2
#č problemem je, že takovéto lineární vyrovnání může davat záporné hodnoty

#č zkusme hledat řešení v mechanice,
#č využijme pružinovou analogii.
#č Máme-li soustavu několika (třeba dvou) pružin, zapojených sériově,
#č kterou zatížíme sílou F, tak deformace soustavy bude:
#č delta_1 + delta_2 = delta = F/k_1 + F/k_2 = F*c_1 + F*c_2
# delat_1 = delta * c_1 / (c_1 + c_2)
# delat_2 = delta * c_2 / (c_1 + c_2)

#č takhle jsme přišli na ekvivalenci nějakého lineárního vyrovnavaní 
#č a sériové soustavy pružin.
#č A teď ten trik - uvažovat deformace těch pružin nelineárně,
#č znemožnit stlačení do záporných délek

# d_epsilon = d_l / l
#č True strain = ln(l/L) = ln(l) - ln(L), kde
#č l je delka pružiny (prutu) po deformaci
#č L je původní délka

#č definujme sílu jako F = strain(epsilon) * tuhost(EA), tj.
# k = EA/l
# poddajnost(c) = původní_délka(l) / EA
# EA = l / c
#č EA vnímáme jako přirozenou charakteristiku, vlastnost
#č u těch charakteristik tuhostí-poddajností 
#č lze všimnout jistou analogii s betama (nebo s CoV),
#č ale liší se jednotky, dělíme rozptílem, nikoliv směrodatnou odchylkou!
#č nebo epsilon = F * poddajnost(c) / původní_délka(l)
#č (výpočet za použití poddajností je hodně jednodušší a přehlednější)
#č řešíme soustavu rovnic
# l_1 + l_2 + l_100500 = L 
# F_1 = F_2 = F_100500 = F
#č zde L je délka soustavy
#č F je síla, působicí na soustavu


def get_true_strain_spring_solution(lenghts, flexibilities, L=1):
    """calculates corrected data for bounded statistically estimated values. 
    It uses nonlinear spring analogy, based on logarithmic strain ε,
     also called, true strain or Hencky strain"""
    #č tohle musí hlídat volající kód
    assert np.all(np.isfinite(lenghts)) and np.all(lenghts > 0)
    assert np.all(np.isfinite(flexibilities)) and np.all(flexibilities > 0)
    
    Ls = lenghts #č původní
    ls = lenghts #č vystupní
    cs = flexibilities
    #č EA vnímáme jako přirozenou charakteristiku, vlastnost
    #č EA jsou z hlediska statistiky vahami
    EAs = lenghts / flexibilities 
    #L = L
    
    i = 0 #č počet iterací
    sum_ls = np.sum(ls)
    
    
    
    #č V tlaku - (záporný) přírůstek síly může (ale nemusí!) narůstat
    #č v tahu Newtonová metoda může nejdřív soustavu 
    #č roztahnout do nekonečna a teprve pak stlačovat zpatky
    #č Právě kvůli tomu v tahu můžou vylezt nekonečna
    #č Jistě máme epsilon bound, definovaný jako delta_L / min(lenghts):
    #
    # linear_epsilon_max = delta_L / min(lenghts),
    #
    #č ale nemůžeme jej použit přímo pro výpočet síly,
    #č F_max vždy vyjde větší jak samotné F-ko za normálního výpočtu:
    #
    # F = delta_L / np.sum(cs)
    # F_max = max(delta_L / flexibilities)
    # F_supermax = linear_epsilon_max * max(EAs)
    # F_supermax = delta_L / min(lenghts) * max(lenghts / flexibilities)
    # F <= F_max <= F_supermax
    #
    
    
    # let's try first approximation
    #č Mám dojem, že se známenko může změnit pouze po první iteraci
    #č Zde může být jak tlak, tak i tah, dále - pouze tlak
    if sum_ls != L:
        #č zaporná - zkracovat, kladná - prodlužovat
        delta_L = L - sum_ls
        # linear approximation
        #č odhadneme sílu
        F = delta_L / np.sum(cs)
        
        #č aha. To je právě ta věc.
        #č nemůžeme rovnou spočítat délky
        #č musíme jít bočnou cestou přes true strain
        # delta_i = F * c_i
        #ls += F * c #ё нихрена подобного!
        
        #č epsilon = F * poddajnost(c) / původní_délka(l)
        epsilons = F / EAs
        ls = np.exp(epsilons) * Ls
        
        i += 1 #č počet iterací
        sum_ls = np.sum(ls)
        print("spring:", i, "delta", delta_L, "F", F)
    
    #č nechcu explicitně výjadřovat přesnost
    #č raději budu sledovat vyši přírůstků.
    #č neplatí, že první iterace musí být větší jak ty další
    #č V tlaku - (záporný) přírůstek síly může (ale nemusí!) narůstat
    #č Doufám, že žádná nekonečna, nanky zde nevylezou
    delta_F = -1 #č ať projdeme kontrolou
    while (sum_ls != L) and (delta_F < 0): # while 1 + 1e-20 != 1
        #č zaporná - zkracovat, kladná - prodlužovat
        delta_L = L - sum_ls
        
        cs = ls / EAs
        # linear approximation
        #č teď bacha, sílu musíme počítat v přirůstcích
        delta_F = delta_L / np.sum(cs)
        F += delta_F
        
        #č epsilon = F * poddajnost(c) / původní_délka(l)
        epsilons = F / EAs
        ls = np.exp(epsilons) * Ls

        i += 1 #č počet iterací
        sum_ls = np.sum(ls)
        print("spring:", i, "delta", delta_L, "F", F)
    
    return ls
        
        
        
def get_semi_true_strain_spring_solution(lenghts, flexibilities, L=1):
    """calculates corrected data for bounded statistically estimated values. 
    It uses nonlinear spring analogy, based on true strain in compression 
    and engineering (Cauchy) strain in tension"""
    
    sum_ls = np.sum(lenghts)
    if sum_ls == L:
        print("spring: no action needed")
        return lenghts
    
    #č tohle musí hlídat volající kód
    assert np.all(np.isfinite(lenghts)) and np.all(lenghts > 0)
    assert np.all(np.isfinite(flexibilities)) and np.all(flexibilities > 0)
    
    Ls = lenghts #č původní
    ls = lenghts #č vystupní
    cs = flexibilities
    
    
    #č v tahu Newtonová metoda mohla by nejdřív soustavu 
    #č roztahnout do nekonečna a teprve pak stlačovat zpatky
    #č Právě kvůli tomu v tahu mohly by vylezt nekonečna
    #
    #č Zde pointa je v tom, že nepotřebujeme precizní pružiny,
    #č postačí nám, když tlak budeme omezovat, 
    #č aby nám délky neujely do záporných hodnot,
    #č zatímco v tahu jednoduše použit lineárné řešení
    
    #č zaporná - zkracovat, kladná - prodlužovat
    delta_L = L - sum_ls
    # linear approximation
    #č odhadneme sílu
    F = delta_L / np.sum(cs)
    
    if F > 0:  #č je třeba roztahovat
        ls += F * cs
        print("spring: tension; linear solution is used.", "Delta=", delta_L)
        return ls
    
    else: #compression, prepare for iteration
        #č EA vnímáme jako přirozenou charakteristiku, vlastnost
        #č EA jsou z hlediska statistiky vahami
        EAs = lenghts / flexibilities 
        
        #č epsilon = F * poddajnost(c) / původní_délka(l)
        epsilons = F / EAs
        ls = np.exp(epsilons) * Ls
        
        i = 1 #č počet iterací
        sum_ls = np.sum(ls)
        print("spring: compression #", i, "delta=", delta_L, "F=", F)
    
    #č nechcu explicitně výjadřovat přesnost
    #č raději budu sledovat vyši přírůstků.
    #č neplatí, že první iterace musí být větší jak ty další
    #č V tlaku - (záporný) přírůstek síly může (ale nemusí!) narůstat
    #č Mám dojem, že se známenko může změnit pouze po první iteraci
    #č Doufám, že žádná nekonečna, nanky zde nevylezou
    delta_F = - np.inf #č ať projdeme kontrolou
    while (sum_ls != L) and (F + delta_F < F): # while 1 + 1e-20 != 1
        #č zaporná - zkracovat, kladná - prodlužovat
        delta_L = L - sum_ls
        
        cs = ls / EAs
        # linear approximation
        #č teď bacha, sílu musíme počítat v přirůstcích
        delta_F = delta_L / np.sum(cs)
        F += delta_F
        
        #č epsilon = F * poddajnost(c) / původní_délka(l)
        epsilons = F / EAs
        ls = np.exp(epsilons) * Ls

        i += 1 #č počet iterací
        sum_ls = np.sum(ls)
        print("spring: compression #", i, "delta=", delta_L, "F=", F)
    
    return ls 
        
        
#č je to implicitní volba     
get_spring_solution = get_semi_true_strain_spring_solution
        
        
#č Mohli bychom počítat v tlaku Eulerem, v tahu - Greenem. 
#оӵ Кинлы со меда кулэ?
#
#def get_euler_almansi_strain_spring_solution(lenghts, flexibilities, L=1):
#    """calculates corrected data for bounded statistically estimated values. 
#    It uses nonlinear spring analogy, based on Euler-Almansi strain"""
#    #č tohle musí hlídat volající kód
#    assert np.all(np.isfinite(lenghts)) and np.all(lenghts > 0)
#    assert np.all(np.isfinite(flexibilities)) and np.all(flexibilities > 0)
#    
#    Ls = lenghts #č původní
#    ls = lenghts #č vystupní
#    cs = flexibilities
#    #č EA vnímáme jako přirozenou charakteristiku, vlastnost
#    #č EA jsou z hlediska statistiky vahami
#    EAs = lenghts / flexibilities 
#    #L = L
#    
#    i = 0 #č počet iterací
#    sum_ls = np.sum(ls)
#    
#    
#    # Euler-Almansi strain
#    # epsilon_EA = (l**2 - L**2) / 2 / l**2
#    
#    
#    #č V tlaku - (záporný) přírůstek síly může (ale nemusí!) narůstat
#    #č v tahu Newtonová metoda může nejdřív soustavu 
#    #č roztahnout do nekonečna a teprve pak stlačovat zpatky
#    #č Právě kvůli tomu v tahu můžou vylezt nekonečna
#    #č Jistě máme epsilon bound, definovaný jako delta_L / min(lenghts):
#    #
#    # linear_epsilon_max = delta_L / min(lenghts),
#    #
#    #č ale nemůžeme jej použit přímo pro výpočet síly,
#    #č F_max vždy vyjde větší jak samotné F-ko za normálního výpočtu:
#    #
#    # F = delta_L / np.sum(cs)
#    # F_max = max(delta_L / flexibilities)
#    # F_supermax = linear_epsilon_max * max(EAs)
#    # F_supermax = delta_L / min(lenghts) * max(lenghts / flexibilities)
#    # F <= F_max <= F_supermax
#    #
#    #č 
#    
#    while (sum_ls < L) and (delta_F < 0): #č je třeba roztahovat
#        #č zaporná - zkracovat, kladná - prodlužovat
#        #č takže delta_L musí být kladná
#        delta_L = L - sum_ls
#        
#        cs = ls / EAs
#        # linear approximation
#        #č teď bacha, sílu musíme počítat v přirůstcích
#        delta_F = delta_L / np.sum(cs)
#        F += delta_F
#        
#        #č epsilon = F * poddajnost(c) / původní_délka(l)
#        epsilons = F / EAs
#        ls = np.exp(epsilons) * Ls
#
#        i += 1 #č počet iterací
#        sum_ls = np.sum(ls)
#        print("spring:", i, "delta", delta_L, "F", F)
#    
#    
#    # let's try first approximation
#    if sum_ls != L:
#        #č zaporná - zkracovat, kladná - prodlužovat
#        delta_L = L - sum_ls
#        # linear approximation
#        #č odhadneme sílu
#        F = delta_L / np.sum(cs)
#        
#        #č aha. To je právě ta věc.
#        #č nemůžeme rovnou spočítat délky
#        #č musíme jít bočnou cestou přes true strain
#        # delta_i = F * c_i
#        #ls += F * c #ё нихрена подобного!
#        
#        #č epsilon = F * poddajnost(c) / původní_délka(l)
#        epsilons = F / EAs
#        ls = np.exp(epsilons) * Ls
#        
#        i += 1 #č počet iterací
#        sum_ls = np.sum(ls)
#        print("spring:", i, "delta", delta_L, "F", F)
#    
#    #č nechcu explicitně výjadřovat přesnost
#    #č raději budu sledovat vyši přírůstků.
#    #č neplatí, že první iterace musí být větší jak ty další
#    #č V tlaku - (záporný) přírůstek síly může (ale nemusí!) narůstat
#    #č v tahu, tuším, tahle Newtonová metoda může nejdřív soustavu 
#    #č roztahnout do nekonečna a teprve pak stlačovat zpatky
#    #č Mám dojem, že se známenko může změnit pouze po první iteraci
#    #č Doufám, že žádná nekonečna, nanky zde nevylezou
#    delta_F = -1 #č ať projdeme kontrolou
#    while (sum_ls != L) and (delta_F < 0): # while 1 + 1e-20 != 1
#        #č zaporná - zkracovat, kladná - prodlužovat
#        delta_L = L - sum_ls
#        
#        cs = ls / EAs
#        # linear approximation
#        #č teď bacha, sílu musíme počítat v přirůstcích
#        delta_F = delta_L / np.sum(cs)
#        F += delta_F
#        
#        #č epsilon = F * poddajnost(c) / původní_délka(l)
#        epsilons = F / EAs
#        ls = np.exp(epsilons) * Ls
#
#        i += 1 #č počet iterací
#        sum_ls = np.sum(ls)
#        print("spring:", i, "delta", delta_L, "F", F)
#    
#    return ls
        

